import sqlite3
import uuid
from itertools import combinations
from os.path import join

import numpy as np
import pandas as pd
from scipy.spatial import Delaunay

from aequilibrae.matrix import AequilibraeMatrix
from aequilibrae.paths import Graph, TrafficClass, TrafficAssignment
from aequilibrae.utils.db_utils import commit_and_close

DELAUNAY_TABLE = "delaunay_network"


class DelaunayAnalysis:
    def __init__(self, project):
        """Start a Delaunay analysis

        :Arguments:
            **project** (:obj:`Project`): The Project to connect to
        """

        self.project = project
        self.procedure_id = uuid.uuid4().hex

    def create_network(self, source="zones", overwrite=False):
        """Creates a delaunay network based on the existing model

        :Arguments:
            **source** (:obj:`str`, *Optional*): Source of the centroids/zones. Either ``zones`` or ``network``. Default ``zones``

            **overwrite path** (:obj:`bool`, *Optional*): Whether to should overwrite an existing Delaunay Network. Default ``False``

        """

        if source not in ["zones", "network"]:
            raise ValueError("Source must be 'zones' or 'network'")

        with self.project.db_connection as conn:
            tables = pd.read_sql("SELECT name FROM sqlite_master WHERE type ='table'", conn)
            if DELAUNAY_TABLE in tables.name.values:
                if not overwrite:
                    raise ValueError("Delaunay network already exist. Use the overwrite flag to re-run it")
                conn.execute(f"DROP TABLE {DELAUNAY_TABLE}")
                conn.execute("delete from geometry_columns where f_table_name=?", [DELAUNAY_TABLE])
                conn.commit()

            zone_sql = "select zone_id node_id, X(st_centroid(geometry)) x, Y(st_centroid(geometry)) y from zones"
            network_sql = "select node_id, X(geometry) x, Y(geometry) y from nodes where is_centroid=1"

            points = pd.read_sql(zone_sql, conn) if source == "zones" else pd.read_sql(network_sql, conn)
            dpoints = np.array(points[["x", "y"]])
            all_edges = Delaunay(np.array(dpoints)).simplices
            edges = []
            for triangle in all_edges:
                links = list(combinations(triangle, 2))
                for i in links:
                    f, t = sorted(i)
                    edges.append([points.at[f, "node_id"], points.at[t, "node_id"]])

            edges = pd.DataFrame(edges)
            edges.drop_duplicates(inplace=True)
            edges.columns = ["a_node", "b_node"]
            edges = edges.assign(direction=0, distance=0, link_id=np.arange(edges.shape[0]) + 1)
            edges = edges[["link_id", "direction", "a_node", "b_node", "distance"]]
            edges.to_sql(DELAUNAY_TABLE, conn, index=False)

            # Now we create the geometries for the delaunay triangulation
            conn.execute(f"select AddGeometryColumn( '{DELAUNAY_TABLE}', 'geometry', 4326, 'LINESTRING', 'XY', 0);")
            conn.execute("CREATE UNIQUE INDEX unique_link_id_delaunay on delaunay_network(link_id)")

            node_geo_sql = """INSERT INTO delaunay_network (link_id, geometry)
                                select lnk.link_id, MakeLine(nd.geometry, nf.geometry) from delaunay_network lnk
                                        inner join nodes nd on lnk.a_node=nd.node_id
                                        inner join nodes nf on lnk.b_node=nf.node_id
                            ON CONFLICT(link_id) DO UPDATE SET geometry=excluded.geometry"""

            zone_geo_sql = """INSERT INTO delaunay_network (link_id, geometry)
                                select lnk.link_id, MakeLine(st_centroid(za.geometry), st_centroid(zb.geometry)) from delaunay_network lnk
                                        inner join zones za on lnk.a_node=za.zone_id
                                        inner join zones zb on lnk.b_node=zb.zone_id
                            ON CONFLICT(link_id) DO UPDATE SET geometry=excluded.geometry;"""

            sql = zone_geo_sql if source == "zones" else node_geo_sql
            conn.execute(sql)
            # Updates link distance
            conn.execute("update delaunay_network set distance=GeodesicLength(geometry);")

    def assign_matrix(self, matrix: AequilibraeMatrix, result_name: str):
        with self.project.db_connection as conn:
            sql = f"select link_id, direction, a_node, b_node, distance, 1 capacity from {DELAUNAY_TABLE}"

            df = pd.read_sql(sql, conn)
            centroids = np.array(np.unique(np.hstack((df.a_node.values, df.b_node.values))), int)

            g = Graph()
            g.mode = "delaunay"
            g.network = df
            g.prepare_graph(centroids)
            g.set_blocked_centroid_flows(True)

            tc = TrafficClass("delaunay", g, matrix)
            ta = TrafficAssignment(self.project)
            ta.set_classes([tc])
            ta.set_time_field("distance")
            ta.set_capacity_field("capacity")
            ta.set_vdf("BPR")
            ta.set_vdf_parameters({"alpha": 0, "beta": 1.0})
            ta.set_algorithm("all-or-nothing")
            ta.execute()

            report = {"setup": str(ta.info())}
            data = [result_name, "Delaunay assignment", self.procedure_id, str(report), ta.procedure_date, ""]
            conn.execute(
                """Insert into results(table_name, procedure, procedure_id, procedure_report, timestamp,
                                                description) Values(?,?,?,?,?,?)""",
                data,
            )

        cols = []
        for x in matrix.view_names:
            cols.extend([f"{x}_ab", f"{x}_ba", f"{x}_tot"])
        df = ta.results()[cols]

        res_path = join(self.project.project_base_path, "results_database.sqlite")
        with commit_and_close(res_path, missing_ok=True) as conn:
            df.to_sql(result_name, conn)



import warnings
from copy import deepcopy
from typing import Union, List, Tuple, Dict
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd

from aequilibrae.matrix import AequilibraeMatrix
from aequilibrae.paths.graph import Graph, TransitGraph, GraphBase
from aequilibrae.paths.results import AssignmentResults, TransitAssignmentResults


class TransportClassBase(ABC):  # noqa: B024
    def __init__(self, name: str, graph: GraphBase, matrix: AequilibraeMatrix) -> None:
        """
        Instantiates the class

        :Arguments:
            **name** (:obj:`str`): UNIQUE class name.

            **graph** (:obj:`Graph`): Class/mode-specific graph

            **matrix** (:obj:`AequilibraeMatrix`): Class/mode-specific matrix. Supports multiple user classes
        """
        if not np.array_equal(matrix.index, graph.centroids):
            raise ValueError("Matrix and graph do not have compatible sets of centroids.")

        if matrix.matrix_view.dtype != graph.default_types("float"):
            raise TypeError("Matrix's computational view need to be of type np.float64")
        self._config = {}
        self.graph = graph
        self.logger = graph.logger
        self.matrix = matrix
        self._id = name

        graph_config = {
            "Mode": graph.mode,
            "Block through centroids": graph.block_centroid_flows,
            "Number of centroids": graph.num_zones,
            "Links": graph.num_links,
            "Nodes": graph.num_nodes,
        }
        self._config["Graph"] = str(graph_config)

        mat_config = {
            "Source": matrix.file_path or "",
            "Number of centroids": matrix.zones,
            "Matrix cores": matrix.view_names,
        }
        if len(matrix.view_names) == 1:
            mat_config["Matrix totals"] = {
                nm: np.sum(np.nan_to_num(matrix.matrix_view)[:, :]) for nm in matrix.view_names
            }
        else:
            mat_config["Matrix totals"] = {
                nm: np.sum(np.nan_to_num(matrix.matrix_view)[:, :, i]) for i, nm in enumerate(matrix.view_names)
            }
        self._config["Matrix"] = str(mat_config)

    @property
    def info(self) -> dict:
        config = deepcopy(self._config)
        return {self._id: config}


class TrafficClass(TransportClassBase):
    """Traffic class for equilibrium traffic assignment

    .. code-block:: python

        >>> from aequilibrae.paths import TrafficClass

        >>> project = create_example(project_path)
        >>> project.network.build_graphs()

        >>> graph = project.network.graphs['c'] # we grab the graph for cars
        >>> graph.set_graph('free_flow_time') # let's say we want to minimize time
        >>> graph.set_skimming(['free_flow_time', 'distance']) # And will skim time and distance
        >>> graph.set_blocked_centroid_flows(True)

        >>> proj_matrices = project.matrices

        >>> demand = proj_matrices.get_matrix("demand_omx")
        >>> demand.computational_view(['matrix'])

        >>> tc = TrafficClass("car", graph, demand)
        >>> tc.set_pce(1.3)
    """

    def __init__(self, name: str, graph: Graph, matrix: AequilibraeMatrix) -> None:
        """
        Instantiates the class

        :Arguments:
            **name** (:obj:`str`): UNIQUE class name.

            **graph** (:obj:`Graph`): Class/mode-specific graph

            **matrix** (:obj:`AequilibraeMatrix`): Class/mode-specific matrix. Supports multiple user classes
        """
        super().__init__(name, graph, matrix)
        self.pce = 1.0
        self.vot = 1.0
        self.mode = graph.mode
        self.class_flow: np.array  # FIXME: Is this ever access?
        self.fixed_cost = np.zeros(graph.graph.shape[0], graph.default_types("float"))
        self.fixed_cost_field = ""
        self.fc_multiplier = 1.0
        self.results = AssignmentResults()
        self._aon_results = AssignmentResults()
        self._selected_links = {}  # maps human name to link_set
        self.congested_time = np.array([])

    def set_pce(self, pce: Union[float, int]) -> None:
        """Sets Passenger Car equivalent

        :Arguments:
            **pce** (:obj:`Union[float, int]`): PCE. Defaults to 1 if not set
        """
        if not isinstance(pce, (float, int)):
            raise ValueError("PCE needs to be either integer or float ")
        self.pce = pce

    def set_fixed_cost(self, field_name: str, multiplier=1):
        """Sets value of time

        :Arguments:
             **field_name** (:obj:`str`): Name of the graph field with fixed costs for this class

             **multiplier** (:obj:`Union[float, int]`): Multiplier for the fixed cost. Defaults to 1 if not set
        """
        if field_name not in self.graph.graph.columns:
            raise ValueError("Field does not exist in the graph")

        self.fc_multiplier = float(multiplier)
        self.fixed_cost_field = field_name
        if np.any(np.isnan(self.graph.graph[field_name].values)):
            self.logger.warning(f"Cost field {field_name} has NaN values. Converted to zero")

        if self.graph.graph[field_name].min() < 0:
            msg = f"Cost field {field_name} has negative values. That is not allowed"
            self.logger.error(msg)
            raise ValueError(msg)

    def set_vot(self, value_of_time: float) -> None:
        """Sets value of time

        :Arguments:
            **value_of_time** (:obj:`Union[float, int]`): Value of time. Defaults to 1 if not set
        """

        self.vot = float(value_of_time)

    def set_select_links(self, links: Dict[str, List[Tuple[int, int]]]):
        """Set the selected links. Checks if the links and directions are valid. Translates link_id and
        direction into unique link id used in compact graph.
        Supply links=None to disable select link analysis.

        :Arguments:
            **links** (:obj:`Union[None, Dict[str, List[Tuple[int, int]]]]`): name of link set and
            Link IDs and directions to be used in select link analysis"""
        self._selected_links = {}
        for name, link_set in links.items():
            if len(name.split(" ")) != 1:
                warnings.warn("Input string name has a space in it. Replacing with _")
                name = str.join("_", name.split(" "))

            link_ids = []
            for link, dir in link_set:
                if dir == 0:
                    query = (self.graph.graph["link_id"] == link) & (
                        (self.graph.graph["direction"] == -1) | (self.graph.graph["direction"] == 1)
                    )
                else:
                    query = (self.graph.graph["link_id"] == link) & (self.graph.graph["direction"] == dir)
                if not query.any():
                    raise ValueError(f"link_id or direction {(link, dir)} is not present within graph.")
                    # Check for duplicate compressed link ids in the current link set
                for comp_id in self.graph.graph[query]["__compressed_id__"].values:
                    if comp_id in link_ids:
                        warnings.warn(
                            "Two input links map to the same compressed link in the network"
                            f", removing superfluous link {link} and direction {dir} with compressed id {comp_id}"
                        )
                    else:
                        link_ids.append(comp_id)
            self._selected_links[name] = np.array(link_ids, dtype=self.graph.default_types("int"))
        self._config["select_links"] = str(links)

    def skim_congested(self, skim_fields=None):
        """
        Skims the congested network. The user can add a list of skims to be computed, which
        will be added to the congested time and the assignment cost from the last iteration of
        the assignment.

        :Arguments:
            **skim_fields** (:obj:`Union[None, str]`): Name of the skims to use. If None, uses default only
        """
        self.graph.graph = self.graph.graph.assign(
            __assignment_cost__=self.graph.cost, __congested_time__=self.congested_time
        )
        skims = (skim_fields or []) + ["__assignment_cost__", "__congested_time__"]
        pre_fields = self.graph.skim_fields
        self.graph.set_skimming(skims)
        skimmer = self.graph.compute_skims()
        self.graph.set_skimming(pre_fields)
        return skimmer

    def __setattr__(self, key, value):
        if key not in [
            "graph",
            "logger",
            "matrix",
            "pce",
            "mode",
            "class_flow",
            "results",
            "_aon_results",
            "_id",
            "vot",
            "fixed_cost",
            "fc_multiplier",
            "fixed_cost_field",
            "_selected_links",
            "_config",
            "congested_time",
        ]:
            raise KeyError(f"Traffic Class does not have '{key}'")
        self.__dict__[key] = value


class TransitClass(TransportClassBase):
    def __init__(self, name: str, graph: TransitGraph, matrix: AequilibraeMatrix):
        super().__init__(name, graph, matrix)
        self._config["Graph"] = str(graph._config)
        self.results = TransitAssignmentResults()

        if len(matrix.view_names) == 1:
            self.matrix_core = matrix.view_names[0]
        else:
            self.matrix_core = None

    def set_demand_matrix_core(self, core: str):
        """
        Set the matrix core to use for demand.

        :Arguments:
            **core** (:obj:`str`):"""
        if core not in self.matrix.view_names:
            raise KeyError(f"'{core}' is not present in `matrix.view_names`")
        self.matrix_core = core



# cython: language_level=3

import multiprocessing
import socket
from pathlib import Path

import pandas as pd

from aequilibrae.context import get_active_project
from aequilibrae.matrix import AequilibraeMatrix
from aequilibrae.project.database_connection import database_path
from aequilibrae.utils.db_utils import commit_and_close

include 'hyperpath.pyx'

from typing import Union


class HyperpathGenerating:
    """A class for hyperpath generation.

    :Arguments:
        **edges** (:obj:`pandas.DataFrame`): The edges of the graph.

        **tail** (:obj:`str`): The column name for the tail of the edge (*Optional*, default is "tail").

        **head** (:obj:`str`): The column name for the head of the edge (*Optional*, default is "head").

        **trav_time** (:obj:`str`): The column name for the travel time of the edge
        (*Optional*, default is "trav_time").

        **freq** (:obj:`str`): The column name for the frequency of the edge (*Optional*, default is "freq").

        **check_edges** (:obj:`bool`): If True, check the validity of the edges (*Optional*, default is False).
    """

    def __init__(
            self,
            edges,
            tail="tail",
            head="head",
            trav_time="trav_time",
            freq="freq",
            check_edges=False,
            skim_cols = None,
            centroids = None
    ):
        skim_cols = self.check_skim_cols(skim_cols)

        # Take a copy as to not modify the users dataframe
        edges, skim_cols = self.compute_skim_cols(skim_cols, edges.copy(), trav_time)
        # load the edges
        if check_edges:
            self._check_edges(edges, tail, head, trav_time, freq, skim_cols)
        self._edges = edges[list(set([tail, head, trav_time, freq] + skim_cols))].copy(deep=True)
        self.edge_count = len(self._edges)

        # remove inf values if any, and values close to zero
        self._edges[trav_time] = np.where(
            self._edges[trav_time] > DATATYPE_INF_PY, DATATYPE_INF_PY, self._edges[trav_time]
        )
        self._edges[trav_time] = np.where(
            self._edges[trav_time] < A_VERY_SMALL_TIME_INTERVAL_PY,
            A_VERY_SMALL_TIME_INTERVAL_PY,
            self._edges[trav_time],
        )
        self._edges[freq] = np.where(self._edges[freq] > INF_FREQ_PY, INF_FREQ_PY, self._edges[freq])
        self._edges[freq] = np.where(self._edges[freq] < MIN_FREQ_PY, MIN_FREQ_PY, self._edges[freq])

        # create an edge index column
        self._edges = self._edges.reset_index(drop=True)
        data_col = "edge_idx"
        self._edges[data_col] = self._edges.index

        # convert to CSC format
        self.vertex_count = self._edges[[tail, head]].max().max() + 1
        rs_indptr, _, rs_data = convert_graph_to_csc_uint32(self._edges, tail, head, data_col, self.vertex_count)
        self._indptr = rs_indptr.astype(np.uint32)
        self._edge_idx = rs_data.astype(np.uint32)

        # edge attributes
        self._trav_time = self._edges[trav_time].values.astype(DATATYPE_PY)
        self._freq = self._edges[freq].values.astype(DATATYPE_PY)
        self._tail = self._edges[tail].values.astype(np.uint32)
        self._head = self._edges[head].values.astype(np.uint32)

        if self._skimming:

            self._is_travel_time = trav_time in skim_cols
            if self._is_travel_time:
                skim_cols.remove(trav_time)
            if not skim_cols:
                self._skim_cols = np.zeros((self._trav_time.shape[0], 1), dtype=DATATYPE_PY)
            else:
                self._skim_cols = self._edges.loc[:, skim_cols].values.astype(DATATYPE_PY)
                self._skim_cols = self._skim_cols.copy(order='C')

            self._skim_cols_names = skim_cols
            self._trav_time_name = trav_time

            # This index already exists, graph.nodes_to_indices. What's the difference between idx and idx_pos?
            self._centroids = centroids.astype(np.uint32)
            self._centroids_idx = np.array(range(len(self._centroids)))
            self._centroids_idx_pos = np.zeros_like(np.array(list(range(self._centroids[-1]+1))))
            for i in self._centroids_idx:
                self._centroids_idx_pos[self._centroids[i]] = i
            self._centroids_idx_pos = self._centroids_idx_pos.astype(np.uint32)

        else:
            self._skim_cols = np.zeros((self._trav_time.shape[0], 1), dtype=DATATYPE_PY)

            self._skim_cols_names = []

            self._centroids = np.array([0], dtype=np.uint32)
            self._centroids_idx_pos = np.array([0], dtype=np.uint32)

            self._is_travel_time = False

    def compute_skim_cols(self, skim_cols, edges: pd.DataFrame, trav_time: str):
        self._get_waiting_time = False
        edges_cols = set(edges.columns)
        skim_cols = set(skim_cols)

        discerete_link_types = {
            "boardings": "boarding",
            "alightings": "alighting",
            "inner_transfers": "inner_transfer",
            "outer_transfers": "outer_transfer",
            "transfers": ["inner_transfer", "outer_transfer"],
        }

        contig_link_types = {
            "on_board_trav_time": "on-board",
            "dwelling_time": "dwell",
            "egress_trav_time": "egress_connector",
            "access_trav_time": "access_connector",
            "walking_trav_time": "walking",
            "transfer_time": ["inner_transfer", "outer_transfer"],
            "in_vehicle_trav_time": ["on-board", "dwell"],
        }

        if any(
                item in skim_cols for item in discerete_link_types.keys() | contig_link_types.keys()
        ) and "link_type" not in edges_cols:
            raise ValueError("predefined skimming type requested but 'link_type' column not present on the graph")

        for name, col in discerete_link_types.items():
            if name in skim_cols and name not in edges_cols:
                if isinstance(col, list):
                    edges[name] = np.where(edges["link_type"].isin(col), 1, 0)
                else:
                    edges[name] = np.where(edges["link_type"] == col, 1, 0)

        if "waiting_time" in skim_cols and "waiting_time" not in edges_cols:
            skim_cols.remove("waiting_time")
            skim_cols = skim_cols | set([trav_time, "in_vehicle_trav_time", "egress_trav_time", "access_trav_time"])
            self._get_waiting_time = True

        for name, col in contig_link_types.items():
            if name in skim_cols and name not in edges_cols:
                if isinstance(col, list):
                    edges[name] = np.where(edges["link_type"].isin(col), edges[trav_time], 0)
                else:
                    edges[name] = np.where(edges["link_type"] == col, edges[trav_time], 0)

        return edges, list(skim_cols)

    def check_skim_cols(self, skim_cols: Union(list[str], tuple[str], set(str))):
        self._skimming = True
        if isinstance(skim_cols, (tuple, set)):
            skim_cols = list(skim_cols)

        if not skim_cols or not isinstance(skim_cols, list):
            skim_cols = []
            self._skimming = False

        return skim_cols

    def _update_od_values(self, origin_column: np.array, destination_column: np.array, demand_column: np.array):

        # get all O-D combinations between centroids
        n = len(self._centroids)
        origin_values = np.repeat(self._centroids, n)

        destination_values = np.tile(self._centroids, n)

        check_bool = origin_values != destination_values
        centroids_origin_column = origin_values[check_bool]
        centroids_destination_column = destination_values[check_bool]

        not_included = ~np.logical_and(
            np.isin(centroids_origin_column, origin_column),
            np.isin(centroids_destination_column, destination_column)
        )  # the combinations of centroids that are not in

        centroids_demand = np.concatenate((demand_column, np.zeros_like(centroids_origin_column[not_included])))
        centroids_origin_column = np.concatenate((origin_column, centroids_origin_column[not_included]))
        centroids_destination_column = np.concatenate((destination_column, centroids_destination_column[not_included]))

        return centroids_origin_column, centroids_destination_column, centroids_demand

    def run(self, origin, destination, volume):
        self.assign(
            np.array([origin]),
            np.array([destination]),
            np.array([volume]),
            threads=1
        )

    def _check_vertex_idx(self, idx):
        assert isinstance(idx, int)
        assert idx >= 0
        assert idx < self.vertex_count

    def _check_volume(self, v):
        assert isinstance(v, float)
        assert v >= 0.0

    def _check_edges(self, edges, tail, head, trav_time, freq, skim_cols):
        if not isinstance(edges, pd.core.frame.DataFrame):
            raise TypeError("edges should be a pandas DataFrame")

        cols = [tail, head, trav_time, freq] + skim_cols

        for col in cols:
            if col not in edges:
                raise KeyError(f"edge column '{col}' not found in graph edges dataframe")

        if edges[cols].isna().any(axis=None):
            raise ValueError(
                " ".join(
                    [
                        f"edges[[{', '.join(map(str, cols))}]]",
                        "should not have any missing value",
                    ]
                )
            )

        for col in [tail, head]:
            if not pd.api.types.is_integer_dtype(edges[col].dtype):
                raise TypeError(f"column '{col}' should be of integer type")

        for col in [trav_time, freq] + skim_cols:
            if not pd.api.types.is_numeric_dtype(edges[col].dtype):
                raise TypeError(f"column '{col}' should be of numeric type")

            if edges[col].min() < 0.0:
                raise ValueError(f"column '{col}' should be nonnegative")

    def assign(
        self,
        origin_column,
        destination_column,
        demand_column,
        check_demand=False,
        threads=None
    ):
        """
        Assigns demand to the edges of the graph.

        Assumes the ``*_column`` arguments are provided as numpy arrays that form a COO sprase matrix.

        :Arguments:
            **origin_column** (:obj:`np.ndarray`): The column for the origin vertices (*Optional*, default is
              "orig_vert_idx").

            **destination_column** (:obj:`np.ndarray`): The column or the destination vertices (*Optional*, default is
              "dest_vert_idx").

            **demand_column** (:obj:`np.ndarray`): The column for the demand values (*Optional*, default is "demand").

            **check_demand** (:obj:`bool`): If True, check the validity of the demand data (*Optional*, default is
              ``False``).

            **threads** (:obj:`int`):The number of threads to use for computation (*Optional*, default is 0, using all
        available threads).
        """

        if self._skimming:
            origin_column, destination_column, demand_column = self._update_od_values(
                origin_column,
                destination_column,
                demand_column
            )

        self.origin_column = origin_column.astype(np.uint32)
        self.destination_column = destination_column.astype(np.uint32)
        self.demand_column = demand_column.astype(DATATYPE_PY)
        # check the input demand parameter
        if check_demand:
            self._check_demand(origin_column, destination_column, demand_column)

        if threads is None:
            threads = 0  # Default to all threads

        # initialize the column storing the resulting edge volumes
        self._edges["volume"] = 0.0

        # travel time is computed but not saved into an array in the following
        self.u_i_vec = np.zeros(self.vertex_count, dtype=DATATYPE_PY)

        # get the list of all destinations
        destination_vertex_indices = np.unique(self.destination_column)

        n_centroids = self._centroids.shape[0]
        n_skim_cols = len(self._skim_cols_names)
        skim_cols = self._skim_cols_names
        if self._is_travel_time:
            skim_cols = [self._trav_time_name] + skim_cols
            n_skim_cols = n_skim_cols + 1

        self.skim_matrix = np.zeros((n_centroids, n_centroids, n_skim_cols))

        compute_SF_in_parallel(
            self._indptr[:],
            self._edge_idx[:],
            self._trav_time[:],
            self._freq[:],
            self._tail[:],
            self._head[:],
            self.destination_column[:],
            destination_vertex_indices[:],
            self.origin_column[:],
            self.demand_column[:],
            self._edges["volume"].values,
            self.vertex_count,
            self._edges["volume"].shape[0],
            (multiprocessing.cpu_count() if threads < 1 else threads),
            self._skim_cols[:],
            self.u_i_vec,
            self.skim_matrix,
            self._centroids[:],
            self._centroids_idx_pos[:],
            self._skimming,
            self._is_travel_time,
            len(self._skim_cols_names)
        )

        if self._skimming:

            self.skim_matrix = self.skim_matrix.transpose(2, 1, 0)
            arr = self.skim_matrix.copy()
            if self._get_waiting_time:
                skim_matrix_dict = {}
                for i in range(n_skim_cols):

                    skim_matrix_dict[skim_cols[i]] = arr[i]

                skim_matrix_dict['waiting_time'] = (
                    skim_matrix_dict['trav_time']
                    - skim_matrix_dict['in_vehicle_trav_time']
                    - skim_matrix_dict['egress_trav_time']
                    - skim_matrix_dict['access_trav_time']
                )
                skim_cols = skim_cols + ['waiting_time']
                arr = np.concatenate((arr, np.expand_dims(skim_matrix_dict['waiting_time'], axis=0)), axis=0)

            self.skim_matrix = AequilibraeMatrix()
            self.skim_matrix.create_empty(zones=len(self._centroids), matrix_names=skim_cols)
            self.skim_matrix.index = self._centroids
            self.skim_matrix.computational_view()
            self.skim_matrix.matrices[:, :, :] = arr.transpose(1, 2, 0)[:, :, :]

        else:
            self.skim_matrix = None

    def _check_demand(self, origin_column, destination_column, demand_column):
        for col, col_name in zip(
                [origin_column, destination_column, demand_column],
                ["origin", "destination", "demand"]
        ):
            if not isinstance(col, (np.ndarray, np.generic)):
                raise TypeError(f"{col_name} should be a numpy array")

            if np.any(np.isnan(col)):
                raise ValueError(f"{col_name} should not have any missing value")

        for col, col_name in zip([origin_column, destination_column], ["origin", "destination"]):
            if not col.dtype == np.uint32:
                raise TypeError(f"column '{col_name}' should be of np.uint32")

        if not demand_column.dtype == np.float64:
            raise TypeError("demand column should be of np.float64 type")

        if demand_column.min() < 0.0:
            raise ValueError("demand column should be nonnegative")

    def info(self) -> dict:
        info = {
            "Algorithm": "Spiess, Heinz & Florian, Michael - Hyperpath generation",
            "Computer name": socket.gethostname(),
            "Procedure ID": self.procedure_id,
        }

        return info

    def save_results(self, table_name: str, keep_zero_flows=True, project=None) -> None:
        """
        Saves the assignment results to results_database.sqlite

        Method fails if table exists

        :Arguments:
            **table_name** (:obj:`str`): Name of the table to hold this assignment result

            **keep_zero_flows** (:obj:`bool`): Whether we should keep records for zero flows. Defaults to ``True``

            **project** (:obj:`Project`, *Optional*): Project we want to save the results to. Defaults to the active
              project
        """

        df = self._edges
        if not keep_zero_flows:
            df = df[df.volume > 0]

        if not project:
            project = project or get_active_project()
        with commit_and_close(Path(project.project_base_path) / "results_database.sqlite", missing_ok=True) as conn:
            df.to_sql(table_name, conn)

        rep = {"setup": self.info()}
        data = [table_name, "hyperpath assignment", self.procedure_id, str(rep), self.procedure_date, self.description]
        sql = """Insert into results(table_name, procedure, procedure_id, procedure_report, timestamp,
                                                                    description) Values(?,?,?,?,?,?)""",
        with commit_and_close(database_path("transit", project.project_base_path)) as conn:
            conn.execute(sql, data)



"""
Algorithms to **calibrate** synthetic gravity models with power and exponential functions

The procedures implemented in this code are some of those suggested in
Modelling Transport, 4th Edition, Ortuzar and Willumsen, Wiley 2011
"""

from time import perf_counter

import numpy as np
import pandas as pd

from aequilibrae.distribution.gravity_application import GravityApplication, SyntheticGravityModel
from aequilibrae.matrix import AequilibraeMatrix
from aequilibrae.parameters import Parameters


class GravityCalibration:
    """Calibrate a traditional gravity model

    Available deterrence function forms are: 'EXPO', 'POWER' or 'GAMMA'.

    .. code-block:: python

        >>> from aequilibrae.distribution import GravityCalibration

        >>> project = create_example(project_path)

        # We load the demand matrix
        >>> demand = project.matrices.get_matrix("demand_omx")
        >>> demand.computational_view()

        # We load the skim matrix
        >>> skim = project.matrices.get_matrix("skims")
        >>> skim.computational_view(["time_final"])

        >>> args = {"matrix": demand,
        ...         "impedance": skim,
        ...         "row_field": "productions",
        ...         "function": 'expo',
        ...         "nan_as_zero": True}
        >>> gravity = GravityCalibration(**args)

        # Solve and save outputs
        >>> gravity.calibrate()
        >>> gravity.model.save(os.path.join(project_path, 'dist_expo_model.mod'))
    """

    def __init__(self, project=None, **kwargs):
        """
        Instantiates the Gravity calibration problem

        :Arguments:
            **matrix** (:obj:`AequilibraeMatrix`): Seed/base trip matrix

            **impedance** (:obj:`AequilibraeMatrix`): Impedance matrix to be used

            **function** (:obj:`str`): Function name to be calibrated. "EXPO" or "POWER"

            **project** (:obj:`Project`, *Optional*): The Project to connect to.
            By default, uses the currently active project

            **parameters** (:obj:`str`, *Optional*): Convergence parameters. Defaults to those in the parameter file

            **nan_as_zero** (:obj:`bool`, *Optional*): If Nan values should be treated as zero. Defaults to ``True``

        :Results:
            **model** (:obj:`SyntheticGravityModel`): Calibrated model

            **report** (:obj:`list`): Iteration and convergence report

            **error** (:obj:`str`): Error description

        """

        self.project = project
        self.__required_parameters = ["max trip length", "max iterations", "max error"]
        self.parameters = kwargs.get("parameters", self.__get_parameters())

        self.nan_as_zero = kwargs.get("nan_as_zero", False)
        self.matrix = kwargs.get("matrix")  # type: AequilibraeMatrix
        self.impedance = kwargs.get("impedance")  # type: AequilibraeMatrix
        deterrence_function = str(kwargs.get("function", "")).upper()

        if self.nan_as_zero:
            self.matrix = self.matrix.copy(memory_only=True)
            self.impedance = self.impedance.copy(memory_only=True)

        self.result_matrix = None
        self.vectors: pd.DataFrame = pd.DataFrame([])
        self.gap = np.inf

        self.error = None
        self.gravity = None

        self.comput_core = None
        self.impedance_core = None

        self.itera = 0
        self.max_iter = None
        self.max_error = None
        self.gap = np.inf

        self.report = ["  #####    GRAVITY CALIBRATION    #####  ", ""]
        self.report.append("Functional form: " + deterrence_function)
        self.model = SyntheticGravityModel()
        if deterrence_function not in self.model.valid_functions:
            raise ValueError("Function needs to be one of these: " + ", ".join(self.model.valid_functions))
        else:
            self.model.function = deterrence_function

    def __assemble_model(self, b1):
        # NEED TO SET PARAMETERS #
        if self.model.function == "EXPO":
            self.model.beta = float(b1)
        elif self.model.function == "POWER":
            self.model.alpha = float(b1)

    def calibrate(self):
        """Calibrate the model

        Resulting model is in *output* class member
        """
        t = perf_counter()
        # initialize auxiliary variables
        max_cost = self.parameters["max trip length"]
        self.max_iter = self.parameters["max iterations"]
        self.max_error = self.parameters["max error"]

        # Check the inputs
        self.__check_inputs()
        if self.model.function in ["EXPO", "POWER"]:
            # filtering for all costs over limit

            a = 1
            if max_cost > 0:
                a = (self.impedance.matrix_view[:, :] < max_cost).astype(int)

            # weighted average cost
            self.report.append("Iteration: 1")
            cstar = np.nansum(self.impedance.matrix_view[:, :] * self.result_matrix.gravity[:, :] * a) / np.nansum(
                self.result_matrix.gravity[:, :] * a
            )

            b0 = 1 / cstar

            self.__assemble_model(b0)
            c0 = self.__apply_gravity()
            for i in self.gravity.report:
                self.report.append("       " + i)
            self.report.append("")
            self.report.append("")

            bm1 = b0
            bm = b0 * c0 / cstar

            self.report.append("Iteration: 2")
            self.__assemble_model(bm)

            cm = self.__apply_gravity()
            for i in self.gravity.report:
                self.report.append("       " + i)
            self.report.append("Error: " + "{:.2E}".format(float(np.nansum(abs((bm / bm1) - 1)))))
            self.report.append("")
            cm1 = c0

        # While the max iterations has not been reached and the error is still too large
        self.itera = 2
        while self.itera < self.max_iter and self.gap > self.max_error:
            self.report.append("Iteration: " + str(self.itera + 1))
            aux = bm
            bm = ((cstar - cm1) * bm - (cstar - cm) * bm) / (cm - cm1)
            bm1 = aux
            cm1 = cm

            self.__assemble_model(bm1)
            cm = self.__apply_gravity()

            for i in self.gravity.report:
                self.report.append("       " + i)
            self.report.append("Error: " + "{:.2E}".format(float(np.nansum(abs((bm / bm1) - 1)))))
            self.report.append("")

            # compute convergence criteria
            self.gap = abs((bm / bm1) - 1)
            self.itera += 1

        if self.itera == self.max_iter:
            self.report.append(
                "DID NOT CONVERGE. Stopped in  " + str(self.itera) + "  with a global error of " + str(self.gap)
            )
        else:
            self.report.append("Converged in " + str(self.itera) + "  iterations to a global error of " + str(self.gap))
        s = perf_counter() - t
        m, s1 = divmod(s, 60)
        s -= m * 60
        h, m = divmod(m, 60)
        t = "%d:%02d:%2.4f" % (h, m, s)

        self.report.append("Running time: " + t)

    def __check_inputs(self):
        if not isinstance(self.impedance, AequilibraeMatrix):
            raise TypeError("Impedance matrix needs to be an instance of AequilibraEMatrix")

        if not isinstance(self.matrix, AequilibraeMatrix):
            raise TypeError("Observed matrix needs to be an instance of AequilibraEMatrix")

        # Check data dimensions
        if not np.array_equal(self.impedance.index, self.impedance.index):
            raise ValueError("Indices from impedance matrix do not match those from seed matrix")

        # Check if matrices were set for computation
        mats = [(self.matrix, "Observed matrix"), (self.impedance, "Impedance matrix")]
        for matrix, title in mats:
            if matrix.matrix_view is None:
                raise ValueError(f"{title} needs to be set for computation")
            if matrix.matrix_view.ndim > 2:
                raise ValueError(f"{title} computational view needs to be set for a single matrix core")
            if np.nansum(matrix.matrix_view.data) == 0:
                raise ValueError(f"{title} has only zero values")
            if np.nanmin(matrix.matrix_view.data) < 0:
                raise ValueError(f"{title} has negative values")

        # Augment parameters if we happen to have only passed one
        default_parameters = self.__get_parameters()
        for para in self.__required_parameters:
            if para not in self.parameters:
                self.parameters[para] = default_parameters[para]

        # Prepare the data for computation
        self.comput_core = self.matrix.view_names[0]

        self.result_matrix = self.matrix.copy(cores=[self.comput_core], names=["gravity"], memory_only=True)

        self.vectors = pd.DataFrame(
            {"rows": self.matrix.rows()[:], "columns": self.matrix.columns()[:]}, index=self.matrix.index[:]
        )

        self.impedance_core = self.impedance.view_names[0]

    def __apply_gravity(self):
        args = {
            "impedance": self.impedance,
            "vectors": self.vectors,
            "row_field": "rows",
            "column_field": "columns",
            "model": self.model,
            "parameters": self.parameters,
            "nan_as_zero": self.nan_as_zero,
        }

        self.gravity = GravityApplication(self.project, **args)
        self.gravity.apply()
        self.result_matrix = self.gravity.output

        return np.nansum(self.impedance.matrix_view[:, :] * self.result_matrix.gravity[:, :]) / np.nansum(
            self.result_matrix.gravity[:, :]
        )

    def __get_parameters(self):
        par = Parameters().parameters
        para = par["distribution"]["ipf"].copy()
        para.update(par["distribution"]["gravity"])
        return para



from libc.math cimport pow
from cython.parallel import prange


def bpr(congested_times, link_flows, capacity, fftime, alpha, beta, cores):
    cdef int c = cores

    cdef double [:] congested_view = congested_times
    cdef double [:] link_flows_view = link_flows
    cdef double [:] capacity_view = capacity
    cdef double [:] fftime_view = fftime
    cdef double [:] alpha_view = alpha
    cdef double [:] beta_view = beta

    with nogil:
        bpr_cython(congested_view, link_flows_view, capacity_view, fftime_view, alpha_view, beta_view, c)


def delta_bpr(dbpr, link_flows, capacity, fftime, alpha, beta, cores):
    cdef int c = cores

    cdef double [:] dbpr_view = dbpr
    cdef double [:] link_flows_view = link_flows
    cdef double [:] capacity_view = capacity
    cdef double [:] fftime_view = fftime
    cdef double [:] alpha_view = alpha
    cdef double [:] beta_view = beta

    with nogil:
        dbpr_cython(dbpr_view, link_flows_view, capacity_view, fftime_view, alpha_view, beta_view, c)


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void bpr_cython(
    double[:] congested_time,
    double[:] link_flows,
    double [:] capacity,
    double [:] fftime,
    double[:] alpha,
    double [:] beta,
    int cores
) noexcept nogil:
    cdef long long i
    cdef long long l = congested_time.shape[0]

    # TODO: Use prange with use_threads_if when Cython 3.1 is released
    for i in range(l):
        if link_flows[i] > 0:
            congested_time[i] = fftime[i] * (1 + alpha[i] * (pow(link_flows[i] / capacity[i], beta[i])))
        else:
            congested_time[i] = fftime[i]


@cython.wraparound(False)
@cython.embedsignature(True)
@cython.boundscheck(False)
cpdef void dbpr_cython(
    double[:] deltaresult,
    double[:] link_flows,
    double [:] capacity,
    double [:] fftime,
    double[:] alpha,
    double [:] beta,
    int cores
) noexcept nogil:
    cdef long long i
    cdef long long l = deltaresult.shape[0]

    # TODO: Use prange with use_threads_if when Cython 3.1 is released
    for i in range(l):
        if link_flows[i] > 0:
            deltaresult[i] = fftime[i] * (
                alpha[i] * beta[i] * (pow(link_flows[i] / capacity[i], beta[i]-1))
            ) / capacity[i]
        else:
            deltaresult[i] = fftime[i]



from aequilibrae.utils.qgis_utils import inside_qgis

if inside_qgis:
    from qgis.PyQt.QtCore import pyqtSignal, QThread

    class WorkerThread(QThread):
        if inside_qgis:
            jobFinished = pyqtSignal(object)

        def __init__(self, parentThread):
            QThread.__init__(self, parentThread)

        def run(self):
            self.running = True
            success = self.doWork()
            if inside_qgis:
                self.jobFinished.emit(success)

        def stop(self):
            self.running = False

else:

    class WorkerThread:  # type: ignore
        def __init__(self, *arg):
            pass



import geopandas as gpd
import pandas as pd

from aequilibrae.utils.find_table_fields import find_table_fields


def get_table(table_name, conn):
    """
    Selects table from database.

    :Arguments:
         **table_name** (:obj:`str`): desired table name
         **conn** (:obj:`sqlite3.Connection`): database connection
    """

    return pd.read_sql(f"SELECT * FROM {table_name};", con=conn)


def get_geo_table(table_name, conn):
    fields, _, geo_field = find_table_fields(table_name, conn=conn)
    fields = [f'"{x}"' for x in fields]
    keys = ",".join(fields)
    if geo_field is not None:
        keys += ', Hex(ST_AsBinary("geometry")) as geometry'

    sql = f"select {keys} from '{table_name}'"
    if geo_field is None:
        return pd.read_sql_query(sql, conn)
    else:
        return gpd.GeoDataFrame.from_postgis(sql, conn, geom_col="geometry", crs="EPSG:4326")



import sqlite3
from aequilibrae.utils.get_table import get_table

# from polarislib.network.data import DataTableStorage
from aequilibrae.transit.transit_elements import Route


def read_routes(conn: sqlite3.Connection):
    data = get_table("transit_routes", conn).reset_index()

    data.drop(
        columns=["pce", "seated_capacity", "design_capacity", "total_capacity", "number_of_cars", "geo"], inplace=True
    )
    data.rename(
        columns={
            "description": "route_desc",
            "longname": "route_long_name",
            "shortname": "route_short_name",
            "type": "route_type",
        },
        inplace=True,
    )

    return [Route(-1).from_row(dt) for _, dt in data.iterrows()]



import logging
from pathlib import Path


class Log:
    """API entry point to the log file contents

    .. code-block:: python

        >>> project = Project()
        >>> project.new(project_path)

        >>> log = project.log()

        # We get all entries for the log file
        >>> entries = log.contents()

        # Or clear everything (NO UN-DOs)
        >>> log.clear()
    """

    def __init__(self, project_base_path: Path):
        self.log_file_path = project_base_path / "aequilibrae.log"

    def contents(self) -> list:
        """Returns contents of log file

        :Returns:
            **log_contents** (:obj:`list`): List with all entries in the log file
        """

        with open(self.log_file_path, "r") as file:
            return [x.strip() for x in file.readlines()]

    def clear(self):
        """Clears the log file. Use it wisely"""
        with open(self.log_file_path, "w") as _:
            pass


def _setup_logger():
    # CREATE THE GLOBAL LOGGER
    logger = logging.getLogger("aequilibrae")
    logger.setLevel(logging.DEBUG)
    return logger


def get_log_handler(log_file: Path, ensure_file_exists=True):
    """Return a log handler that writes to the given log_file"""
    if log_file.exists() and not log_file.is_file():
        raise FileExistsError(f"{log_file} is not a valid file")

    if ensure_file_exists:
        open(log_file, "a").close()

    formatter = logging.Formatter("%(asctime)s;%(levelname)s ; %(message)s")
    handler = logging.FileHandler(log_file)
    handler.setFormatter(formatter)
    handler.name = "aequilibrae"
    handler.setLevel(logging.DEBUG)
    return handler


global_logger = logger = _setup_logger()



from os.path import join

import pandas as pd

from aequilibrae.parameters import Parameters


class GMNSExporter:
    def __init__(self, net, path) -> None:
        self.p = Parameters()
        self.links_df = net.links.data
        self.nodes_df = net.nodes.data
        self.source = net.project.path_to_file
        self.output_path = path

        self.gmns_parameters = self.p.parameters["network"]["gmns"]
        self.gmns_links = self.gmns_parameters["link"]
        self.gmns_nodes = self.gmns_parameters["node"]

        with net.project.db_connection as conn:
            cur = conn.execute("select mode_name, mode_id, description, pce, vot, ppv from modes").fetchall()
        self.modes_df = pd.DataFrame(cur, columns=["mode_name", "mode_id", "description", "pce", "vot", "ppv"])

    def doWork(self):
        if "ogc_fid" in list(self.links_df.columns):
            self.links_df.drop("ogc_fid", axis=1, inplace=True)
        if "ogc_fid" in list(self.nodes_df.columns):
            self.nodes_df.drop("ogc_fid", axis=1, inplace=True)

        self.update_direction_field()

        # Converting from meters to kilometers
        self.links_df.distance /= 1000

        self.update_field_names()

        self.reorder_fields()

        # Exporting network (links and nodes)
        self.links_df.to_csv(join(self.output_path, "link.csv"), index=False)
        self.nodes_df.to_csv(join(self.output_path, "node.csv"), index=False)

        self.update_modes_fields()

        # Exporting use_definition table
        self.modes_df.to_csv(join(self.output_path, "use_definition.csv"), index=False)

    def update_direction_field(self):
        two_way_cols = list({col[:-3] for col in list(self.links_df.columns) if col[-3:] in ["_ab", "_ba"]})

        ab_links = pd.DataFrame(self.links_df[self.links_df.direction > -1], copy=True)
        ba_links = pd.DataFrame(self.links_df[self.links_df.direction < 1], copy=True)

        # treats ab_links and bi-directionals
        if ab_links.shape[0]:
            ab_links.loc[:, "dir_flag"] = 1
            for col in two_way_cols:
                ab_links.loc[:, col] = ab_links.loc[:, col + "_ab"]

        # treats ba_links and bi-directionals
        if ba_links.shape[0]:
            ba_links.loc[:, "direction"] = 1
            ba_links.loc[:, "dir_flag"] = -1
            b = ba_links.b_node.to_numpy()
            ba_links.loc[:, "b_node"] = ba_links.a_node.to_numpy()[:]
            ba_links.loc[:, "a_node"] = b[:]

            for col in two_way_cols:
                ba_links.loc[:, col] = ba_links.loc[:, col + "_ba"]

        self.links_df = pd.concat([ab_links, ba_links])

    def update_field_names(self):
        """
        Updates field names according to equivalency between AequilibraE and GMNS fields.
        """

        links_equiv = self.gmns_links["equivalency"]
        nodes_equiv = self.gmns_nodes["equivalency"]

        for col in list(self.links_df.columns):
            if col in links_equiv:
                if links_equiv[col] not in list(self.links_df.columns):
                    self.links_df.rename(columns={f"{col}": f"{links_equiv[col]}"}, inplace=True)
                elif col not in ["lanes", "capacity", "link_id", "name", "geometry"]:
                    self.links_df.drop(col, axis=1, inplace=True)
            elif col[-3:] in ["_ab", "_ba"]:
                self.links_df.drop(col, axis=1, inplace=True)

        for idx, row in self.nodes_df.iterrows():
            self.nodes_df.loc[idx, "node_type"] = "centroid" if row.is_centroid == 1 else None

        for col in list(self.nodes_df.columns):
            if col in nodes_equiv:
                if nodes_equiv[col] not in list(self.nodes_df.columns):
                    self.nodes_df.rename(columns={f"{col}": f"{nodes_equiv[col]}"}, inplace=True)
                elif col != "node_id":
                    self.links_df.drop(col, axis=1, inplace=True)
            elif col == "geometry":
                self.nodes_df = self.nodes_df.assign(
                    x_coord=[self.nodes_df.geometry[idx].coords[0][0] for idx in list(self.nodes_df.index)]
                )
                self.nodes_df = self.nodes_df.assign(
                    y_coord=[self.nodes_df.geometry[idx].coords[0][1] for idx in list(self.nodes_df.index)]
                )
                self.nodes_df.drop("geometry", axis=1, inplace=True)

    def reorder_fields(self):
        link_cols = list(self.links_df.columns)
        link_req = [k for k in self.gmns_links["fields"] if self.gmns_links["fields"][k]["required"]]
        main_cols = ["link_id", "from_node_id", "to_node_id", "directed"]
        link_cols = (
            main_cols
            + [c for c in link_cols if c in link_req and c not in main_cols]
            + [c for c in link_cols if c not in link_req]
        )

        node_cols = list(self.nodes_df.columns)
        node_req = [k for k in self.gmns_nodes["fields"] if self.gmns_nodes["fields"][k]["required"]]
        main_cols = ["node_id", "x_coord", "y_coord"]
        node_cols = (
            main_cols
            + [c for c in node_cols if c in node_req and c not in main_cols]
            + [c for c in node_cols if c not in node_req]
        )

        self.links_df = self.links_df[link_cols]
        self.nodes_df = self.nodes_df[node_cols]

    def update_modes_fields(self):
        """
        Updates AequilibraE modes table so it can be exported as a GMNS use_definition table.
        """

        fields_dict = self.gmns_parameters["use_definition"]["equivalency"]
        self.modes_df = self.modes_df[["mode_name", "ppv", "pce", "description", "mode_id"]].rename(
            columns={"mode_name": fields_dict["mode_name"], "ppv": "persons_per_vehicle"}
        )



import sqlite3

import shapely.wkb
from aequilibrae.utils.get_table import get_table

# from polarislib.network.data import DataTableStorage
from aequilibrae.transit.transit_elements import Stop


def read_stops(conn: sqlite3.Connection, transformer):
    data = get_table("transit_stops", conn).reset_index()
    data = data[data.agency_id > 1]
    data.geo = data.geo.apply(shapely.wkb.loads)
    if transformer:
        lons, lats = transformer.transform(data.X.values, data.Y.values)
        data.loc[:, "X"] = lons[:]
        data.loc[:, "Y"] = lats[:]

    data.drop(columns=["moved_by_matching", "Z"], inplace=True)
    data.rename(
        columns={
            "description": "stop_desc",
            "name": "stop_name",
            "street": "stop_street",
            "transit_zone_id": "zone_id",
            "Y": "stop_lat",
            "X": "stop_lon",
        },
        inplace=True,
    )

    return [Stop(-1).from_row(dt) for _, dt in data.iterrows()]



from sqlite3 import IntegrityError

from aequilibrae.project.field_editor import FieldEditor
from aequilibrae.project.network.mode import Mode


class Modes:
    """
    Access to the API resources to manipulate the modes table in the network

    .. code-block:: python

        >>> project = create_example(project_path)

        >>> modes = project.network.modes

        # We can get a dictionary of all modes in the model
        >>> all_modes = modes.all_modes()

        # And do a bulk change and save it
        >>> for mode_id, mode_obj in all_modes.items():
        ...     mode_obj.beta = 1
        ...     mode_obj.save()

        # or just get one mode in specific
        >>> car_mode = modes.get('c')

        # or just get this same mode by name
        >>> car_mode = modes.get_by_name('car')

        # We can change the description of the mode
        >>> car_mode.description = 'personal autos only'

        # Let's say we are using alpha to store the PCE for a future year with much smaller cars
        >>> car_mode.alpha = 0.95

        # To save this mode we can simply
        >>> car_mode.save()

        # We can also create a completely new mode and add to the model
        >>> new_mode = modes.new('k')
        >>> new_mode.mode_name = 'flying_car'  # Only ASCII letters and *_* allowed # other fields are not mandatory

        # We then explicitly add it to the network
        >>> modes.add(new_mode)

        # we can even keep editing and save it directly once we have added it to the project
        >>> new_mode.description = 'this is my new description'
        >>> new_mode.save()
    """

    def __init__(self, net):
        self.__all_modes = []
        self.__items = {}
        self.project = net.project
        self.logger = net.logger
        self.__update_list_of_modes()

    def add(self, mode: Mode) -> None:
        """We add a mode to the project"""
        self.__update_list_of_modes()

        with self.project.db_connection as conn:
            if mode.mode_id in self.__all_modes:
                raise ValueError("Mode already exists in the model")

            conn.execute("insert into 'modes'(mode_id, mode_name) Values(?,?)", [mode.mode_id, mode.mode_name])
            self.logger.info(f"mode {mode.mode_name}({mode.mode_id}) was added to the project")
            conn.commit()
            mode.save()

        self.__update_list_of_modes()

    def delete(self, mode_id: str) -> None:
        """Removes the mode with *mode_id* from the project"""
        with self.project.db_connection as conn:
            try:
                conn.execute(f'delete from modes where mode_id="{mode_id}"')
            except IntegrityError as e:
                self.logger.error(f"Failed to remove mode {mode_id}. {e.args}")
                raise e
            self.logger.warning(f"Mode {mode_id} was successfully removed from the database")
        self.__update_list_of_modes()

    @property
    def fields(self) -> FieldEditor:
        """Returns a FieldEditor class instance to edit the Modes table fields and their metadata"""
        return FieldEditor(self.project, "modes")

    def get(self, mode_id: str) -> Mode:
        """Get a mode from the network by its *mode_id*"""
        self.__update_list_of_modes()
        if mode_id not in self.__all_modes:
            raise ValueError(f"Mode {mode_id} does not exist in the model")
        return Mode(mode_id, self.project)

    def get_by_name(self, mode: str) -> Mode:
        """Get a mode from the network by its *mode_name*"""
        self.__update_list_of_modes()
        with self.project.db_connection as conn:
            found = conn.execute(f"select mode_id from 'modes' where mode_name='{mode}'").fetchone()
        if len(found) == 0:
            raise ValueError(f"Mode {mode} does not exist in the model")
        return Mode(found[0], self.project)

    def all_modes(self) -> dict:
        """Returns a dictionary with all mode objects available in the model. mode_id as key"""
        self.__update_list_of_modes()
        return {x: Mode(x, self.project) for x in self.__all_modes}

    def new(self, mode_id: str) -> Mode:
        """Returns a new mode with *mode_id* that can be added to the model later"""
        if mode_id in self.__all_modes:
            raise ValueError("Mode already exists in the model. Creating a new one does not make sense")

        return Mode(mode_id, self.project)

    def __update_list_of_modes(self) -> None:
        with self.project.db_connection as conn:
            self.__all_modes = [x[0] for x in conn.execute("select mode_id from 'modes'").fetchall()]

    def __copy__(self):
        raise Exception("Modes object cannot be copied")

    def __deepcopy__(self, memodict=None):
        raise Exception("Modes object cannot be copied")

    def __del__(self):
        self.__items.clear()



from aequilibrae.matrix.coo_demand cimport GeneralisedCOODemand
from aequilibrae.paths.cython.route_choice_types cimport (
    RouteVec_t,
    RouteSet_t,
)

from libcpp.vector cimport vector
from libcpp.memory cimport shared_ptr
from libc.stdint cimport *


cdef class RouteChoiceSetResults:
    cdef:
        GeneralisedCOODemand demand
        bint store_results
        bint perform_assignment
        double cutoff_prob
        double beta
        double[:] cost_view
        unsigned int [:] mapping_idx
        int64_t [::] mapping_data

        vector[shared_ptr[RouteVec_t]] __route_vecs
        vector[vector[long long] *] __link_union_set
        vector[shared_ptr[vector[double]]] __cost_set
        vector[shared_ptr[vector[bint]]] __mask_set
        vector[shared_ptr[vector[double]]] __path_overlap_set
        vector[shared_ptr[vector[double]]] __prob_set

        readonly object table

    @staticmethod
    cdef void route_set_to_route_vec(RouteVec_t &route_vec, RouteSet_t &route_set) noexcept nogil

    cdef shared_ptr[RouteVec_t] get_route_vec(RouteChoiceSetResults self, size_t i) noexcept nogil
    cdef shared_ptr[vector[double]] __get_cost_set(RouteChoiceSetResults self, size_t i) noexcept nogil
    cdef shared_ptr[vector[bint]] __get_mask_set(RouteChoiceSetResults self, size_t i) noexcept nogil
    cdef shared_ptr[vector[double]] __get_path_overlap_set(RouteChoiceSetResults self, size_t i) noexcept nogil
    cdef shared_ptr[vector[double]] get_prob_vec(RouteChoiceSetResults self, size_t i) noexcept nogil

    cdef shared_ptr[vector[double]] compute_result(
        RouteChoiceSetResults self,
        size_t i,
        RouteVec_t &route_set,
        bint *found_zero_cost,
        size_t thread_id
    ) noexcept nogil

    cdef void compute_cost(
        RouteChoiceSetResults self,
        vector[double] &cost_vec,
        const RouteVec_t &route_set,
        const double[:] cost_view,
        bint *found_zero_cost
    ) noexcept nogil

    cdef void compute_mask(
        RouteChoiceSetResults self,
        vector[bint] &route_mask,
        const vector[double] &total_cost
    ) noexcept nogil

    cdef void compute_frequency(
        RouteChoiceSetResults self,
        vector[long long] &keys,
        vector[long long] &counts,
        const RouteVec_t &route_set,
        const vector[bint] &route_mask
    ) noexcept nogil

    cdef void compute_path_overlap(
        RouteChoiceSetResults self,
        vector[double] &path_overlap_vec,
        const RouteVec_t &route_set,
        const vector[long long] &keys,
        const vector[long long] &counts,
        const vector[double] &total_cost,
        const vector[bint] &route_mask,
        const double[:] cost_view
    ) noexcept nogil

    cdef void compute_prob(
        RouteChoiceSetResults self,
        vector[double] &prob_vec,
        const vector[double] &total_cost,
        const vector[double] &path_overlap_vec,
        const vector[bint] &route_mask
    ) noexcept nogil

    cdef object make_df_from_results(RouteChoiceSetResults self)

cdef double inverse_binary_logit(double prob, double beta0, double beta1) noexcept nogil



import hashlib
import math
from contextlib import closing
from copy import deepcopy
from os.path import isfile, join
from tempfile import gettempdir

import numpy as np
import pandas as pd
import shapely.wkt
from shapely.geometry import LineString
from shapely.geometry import Point
from shapely.ops import substring

from aequilibrae.log import logger
from aequilibrae.project.database_connection import database_connection
from aequilibrae.project.zoning import GeoIndex
from aequilibrae.transit.constants import DRIVING_SIDE
from aequilibrae.transit.functions.compute_line_bearing import compute_line_bearing
from aequilibrae.transit.transit_elements import mode_correspondence
from aequilibrae.utils.aeq_signal import simple_progress
from aequilibrae.utils.interface.worker_thread import WorkerThread

GRAPH_VERSION = 1
CONNECTOR_SPEED = 1


class MMGraph(WorkerThread):
    """Build specialized map-matching graphs. Not designed to be used by the final user"""

    def __init__(self, lib_gtfs):
        WorkerThread.__init__(self, None)

        self.project = lib_gtfs.project
        self.stops = lib_gtfs.gtfs_data.stops
        self.lib_gtfs = lib_gtfs
        self._idx = None
        self.max_link_id = -1
        self.max_node_id = -1
        self.mode = ""
        self.modename = ""
        self.mode_id = -1
        self.__mode = ""
        self.__df_file = ""
        self.__agency = lib_gtfs.gtfs_data.agency.agency
        self.__centroids_file = ""
        self.__mm_graph_file = ""
        self.node_corresp = []
        self.__all_links = {}
        self.distance_to_project = -1
        self.df = pd.DataFrame([])
        self.logger = logger
        self.signal = lib_gtfs.signal

    def build_graph_with_broken_stops(self, mode_id: int, distance_to_project=200):
        """Build the graph for links for a certain mode while splitting the closest links at stops' projection

        :Arguments:
            **mode_id** (:obj:`int`): Mode ID for which we will build the graph for

            **distance_to_project** (:obj:`float`, *Optional*): Radius search for links to break at the stops.
            Defaults to 50m
        """
        self.logger.debug(f"Called build_graph_with_broken_stops for mode_id={mode_id}")

        self.mode_id = mode_id
        self.distance_to_project = distance_to_project
        self.__mode = mode_correspondence[self.mode_id]
        self.__mm_graph_file = join(gettempdir(), f"map_matching_graph_{self.__agency}_{self.__mode}.csv")
        modename = self.project.network.modes.get(self.__mode).mode_name

        with closing(database_connection("network")) as conn:
            get_qry = f"""Select link_id, a_node, b_node, max(speed_ab, speed_ba) speed,
                          distance, ST_AsText(geometry) wkt from links
                          WHERE INSTR(links.modes, "{self.__mode}")>0 AND direction>=0
                          UNION ALL
                          Select  link_id * -1 , b_node a_node, a_node b_node, max(speed_ab, speed_ba) speed,
                          distance, ST_AsText(ST_Reverse(geometry)) wkt from links
                          WHERE INSTR(links.modes, "{self.__mode}")>0 AND direction<=0;"""

            self.logger.debug("  Reading links from physical network")
            self.df = pd.read_sql(get_qry, conn)

        if not self.df.shape[0]:
            from aequilibrae.paths import Graph

            return Graph()

        # We do some wrangling to save the graph to disk, in case we need to run this more than once
        authvalue = hashlib.md5()
        authvalue.update(str(distance_to_project).encode())
        authvalue.update(str(GRAPH_VERSION).encode())
        authvalue.update("".join([str(x) for x in self.stops.keys()]).encode())
        authvalue.update("".join([str(x) for x in self.df.link_id.values]).encode())
        authvalue.update(modename.encode())
        graph_hash = authvalue.hexdigest()

        self.__df_file = join(gettempdir(), f"{graph_hash}_df.zip")
        self.__centroids_file = join(gettempdir(), f"{graph_hash}_centroids.zip")
        if isfile(self.__df_file) and isfile(self.__centroids_file):
            return self.__build_graph_from_cache()

        return self.__build_graph_from_scratch()

    def __build_graph_from_cache(self):
        msg = f"Loading map-matching graph from disk for mode_id={self.mode_id} (Step: 8/12)"
        self.logger.info(msg)
        self.signal.emit(["set_text", msg])

        net = pd.read_csv(self.__df_file)
        centroid_corresp = pd.read_csv(self.__centroids_file)
        centroids = np.copy(centroid_corresp.centroid_id.values)
        centroid_corresp.set_index("node_id", inplace=True)
        for stop in self.stops.values():
            stop.__map_matching_id__[self.mode_id] = centroid_corresp.loc[stop.stop_id, "centroid_id"]

        return self.__graph_from_broken_net(centroids, net)

    def __build_graph_from_scratch(self):
        msg = f"Creating map-matching graph from scratch for mode_id={self.mode_id} (Step: 8/12)"
        self.logger.info(msg)
        self.signal.emit(["set_text", msg])

        self.df = self.df.assign(original_id=self.df.link_id, is_connector=0, geo=self.df.wkt.apply(shapely.wkt.loads))
        self.df.loc[self.df.link_id < 0, "link_id"] = self.df.link_id * -1 + self.df.link_id.max() + 1

        # We make sure all link IDs are in proper order
        self.max_link_id = self.df.link_id.max() + 1
        self.max_node_id = self.df[["a_node", "b_node"]].max().max() + 1

        # Build initial index
        self._idx = GeoIndex()
        msg = "Building graphs - Indexing links (Step: 8/12)"
        for _, record in simple_progress(self.df.iterrows(), self.signal, msg):
            self._idx.insert(feature_id=record.link_id, geometry=record.geo)

        # We will progressively break links at stops' projection
        # But only on the right side of the link (no boarding at the opposing link's side)
        centroids = []
        self.node_corresp = []
        self.df = self.df.assign(direction=1, free_flow_time=np.inf, wrong_side=0, closest=1, to_remove=0)
        self.__all_links = {rec.link_id: rec for _, rec in self.df.iterrows()}

        msg = "Building graphs - Breaking links (Step: 8/12)"
        for stop_id, stop in simple_progress(self.stops.items(), self.signal, msg):
            stop.__map_matching_id__[self.mode_id] = self.max_node_id
            self.node_corresp.append([stop_id, self.max_node_id])
            centroids.append(stop.__map_matching_id__[self.mode_id])
            self.max_node_id += 1
            self.connect_node(stop)
        self.df = pd.concat([pd.DataFrame(rec).transpose() for rec in self.__all_links.values()])

        self.df = self.df[self.df.to_remove == 0]
        fltr = self.df.speed > 0
        self.df.loc[fltr, "free_flow_time"] = self.df.distance[fltr] / self.df.speed[fltr]

        cols = [
            "link_id",
            "a_node",
            "b_node",
            "direction",
            "free_flow_time",
            "distance",
            "is_connector",
            "closest",
            "original_id",
        ]
        net = self.df[cols]
        # Caches the graph outputs
        net.reset_index(inplace=True, drop=True)
        net = net.drop_duplicates(subset=["a_node", "b_node"])
        net.to_csv(self.__df_file, index=False)
        cols.append("wkt")
        self.df[cols].to_csv(self.__mm_graph_file, index=False)
        pd.DataFrame(self.node_corresp, columns=["node_id", "centroid_id"]).to_csv(self.__centroids_file)
        return self.__graph_from_broken_net(centroids, net)

    def connect_node(self, stop) -> None:
        list_nearest = list(self._idx.nearest(stop.geo, 30))

        is_closest = 1
        conn_found = 0
        distances = []
        for lid in list_nearest:
            lgeo = self.__all_links[lid].geo
            distances.append(stop.geo.distance(lgeo) * math.pi * 6371000 / 180)

        # Sort by distance to the stop
        nearest = pd.DataFrame({"dist": distances, "link_id": list_nearest}).sort_values(by="dist")
        criterium = min(5 * nearest.dist.min(), self.distance_to_project)
        criterium = criterium if nearest.dist.min() < 250 else nearest.dist.min()
        nearest = nearest[nearest.dist <= max(criterium, nearest.dist.min())].link_id.tolist()

        for counter, link_id in enumerate(nearest):
            wrong_side = 0
            link = self.__all_links[link_id]

            link_geo = link.geo  # Linestring

            # We disregard links beyond the threshold, but maintain the closest link to ensure connectivity

            if link_geo.boundary.is_empty:
                first = Point([link_geo.coords.xy[0][0], link_geo.coords.xy[1][0]])
                last = Point([link_geo.coords.xy[0][-1], link_geo.coords.xy[1][-1]])
            else:
                first = link_geo.boundary.geoms[0]
                last = link_geo.boundary.geoms[1]

            proj_point = link_geo.project(stop.geo)
            corr_proj = proj_point * math.pi * 6371000 / 180
            break_point = link_geo.interpolate(proj_point)
            connector_geo = LineString([stop.geo, break_point])
            conn_length = connector_geo.length * math.pi * 6371000 / 180

            if conn_length > 0:
                p = break_point if corr_proj > 0 else last
                az_link = compute_line_bearing((first.x, first.y), (p.x, p.y))
                az_connector = compute_line_bearing((stop.geo.x, stop.geo.y), (break_point.x, break_point.y))
                if (az_link - az_connector) * DRIVING_SIDE < 0:
                    wrong_side = 1  # We are on the WRONG side

            if corr_proj <= 1.0:  # If within one meter of the end of the link, let's go with the existing node
                break_point = first
                intersec_node = link.a_node
            elif corr_proj >= (link_geo.length * math.pi * 6371000 / 180) - 1.0:
                break_point = last
                intersec_node = link.b_node
            else:
                link.to_remove = 1
                self._idx.delete(link_id, link_geo)
                intersec_node = self.max_node_id
                self.max_node_id += 1

                # Create the first portion of the link
                fp = deepcopy(link)
                fp.link_id = self.max_link_id
                fp.b_node = intersec_node
                fp.to_remove = 0
                fp.geo = substring(link_geo, 0, proj_point)
                fp.wkt = fp.geo.wkt
                fp.distance = fp.geo.length * math.pi * 6371000 / 180
                self._idx.insert(fp.link_id, fp.geo)
                self.max_link_id += 1
                self.__all_links[fp.link_id] = fp

                # Create the second portion of the link
                lp = deepcopy(link)
                lp.link_id = self.max_link_id
                lp.a_node = intersec_node
                lp.geo = substring(link_geo, proj_point, link_geo.length)
                lp.wkt = lp.geo.wkt
                lp.distance = lp.geo.length * math.pi * 6371000 / 180
                lp.to_remove = 0
                self._idx.insert(lp.link_id, lp.geo)
                self.max_link_id += 1
                self.__all_links[lp.link_id] = lp

            # Add the connector
            connector_geo = LineString([stop.geo, break_point])
            connector = deepcopy(link)
            connector.link_id = self.max_link_id
            connector.original_id = -1
            connector.a_node = stop.__map_matching_id__[self.mode_id]
            connector.b_node = intersec_node
            connector.wrong_side = wrong_side
            connector.direction = 0
            connector.to_remove = 0
            connector.geo = connector_geo
            connector.wkt = connector_geo.wkt
            connector.distance = connector_geo.length * math.pi * 6371000 / 180
            connector.is_connector = 1  # We make sure that the closest connector cannot be turned off
            connector.closest = is_closest
            self.max_link_id += 1
            is_closest = 0
            conn_found += 1
            self.__all_links[connector.link_id] = connector

    def __graph_from_broken_net(self, centroids, net):
        from aequilibrae.paths import Graph

        net_data = pd.DataFrame(
            {
                "distance": net.distance.astype(float),
                "direction": net.direction.astype(np.int8),
                "a_node": net.a_node.astype(int),
                "b_node": net.b_node.astype(int),
                "link_id": net.link_id.astype(int),
                "is_connector": net.is_connector.astype(int),
                "original_id": net.original_id.astype(int),
                "closest": net.closest.astype(int),
                "free_flow_time": net.free_flow_time.astype(float),
            }
        )

        fltr = net.is_connector == 1
        net_data.loc[fltr, "distance"] = 1.2 * (net_data[fltr].distance ** 1.3)  # Already penalize it a bit
        net_data.loc[fltr, "free_flow_time"] = ((net_data[fltr].distance / 1000) / CONNECTOR_SPEED) * 60

        g = Graph()
        g.network = net_data
        g.prepare_graph(np.array(centroids))
        g.set_graph("free_flow_time")
        g.set_skimming(["distance"])
        g.set_blocked_centroid_flows(True)
        return g



import string


class Mode:
    """A mode object represents a single record in the *modes* table"""

    __alowed_characters = string.ascii_letters + "_"

    def __init__(self, mode_id: str, project) -> None:
        self.project = project
        if mode_id is None:
            raise ValueError("Mode IDs cannot be None")

        if len(mode_id) != 1 or mode_id not in string.ascii_letters:
            raise ValueError("Mode IDs must be a single ascii character")

        with self.project.db_connection as conn:
            table_struct = conn.execute("pragma table_info(modes)").fetchall()
            self.__fields = [x[1] for x in table_struct]
            self.__original__ = {}
            # data for the mode
            dt = conn.execute(f"select * from 'modes' where mode_id='{mode_id}'").fetchone()

        if dt is None:
            # if the mode was not found, we return a new one
            for k in self.__fields:
                self.__dict__[k] = None
                self.__original__[k] = None
            self.__dict__["mode_id"] = mode_id
            self.__original__["mode_id"] = mode_id
        else:
            for k, v in zip(self.__fields, dt):
                self.__dict__[k] = v
                self.__original__[k] = v

    def __setattr__(self, instance, value) -> None:
        if instance == "mode_name" and value is None:
            raise ValueError("mode_name cannot be None")

        if instance == "mode_id":
            raise ValueError("Changing a mode_id is not supported. Create a new one and delete this one")
        else:
            self.__dict__[instance] = value

    def save(self):
        if self.mode_id not in self.__alowed_characters:
            raise ValueError("mode_id needs to be a ascii letter")

        for letter in self.mode_name:
            if letter not in self.__alowed_characters:
                raise ValueError('mode_name can only contain letters and "_"')

        with self.project.db_connection as conn:
            if conn.execute(f'select count(*) from modes where mode_id="{self.mode_id}"').fetchone()[0] == 0:
                raise ValueError("Mode does not exist in the model. You need to explicitly add it")

            table_struct = [x[1] for x in conn.execute("pragma table_info(modes)").fetchall()]

            for key, value in self.__dict__.items():
                if key in table_struct and key != "mode_id":
                    v_old = self.__original__.get(key, None)
                    if value != v_old and value is not None:
                        self.__original__[key] = value
                        conn.execute(f"update 'modes' set '{key}'=? where mode_id='{self.mode_id}'", [value])



import sqlite3

# from aequilibrae.data import DataTableStorage
from aequilibrae.utils.get_table import get_table
from aequilibrae.transit.transit_elements import Agency


def read_agencies(conn: sqlite3.Connection):
    data = get_table("transit_agencies", conn).reset_index()
    return [Agency().from_row(dt) for _, dt in data.iterrows() if dt.agency_id > 1]



import functools
import logging
import os
import shutil
import sqlite3
from contextlib import contextmanager
from pathlib import Path

from aequilibrae import global_logger
from aequilibrae.context import activate_project, get_active_project
from aequilibrae.log import Log
from aequilibrae.log import get_log_handler
from aequilibrae.parameters import Parameters
from aequilibrae.project.about import About
from aequilibrae.project.data import Matrices
from aequilibrae.project.network import Network
from aequilibrae.project.project_cleaning import clean
from aequilibrae.project.project_creation import initialize_tables
from aequilibrae.project.zoning import Zoning
from aequilibrae.project.tools import MigrationManager
from aequilibrae.project.database_connection import database_connection
from aequilibrae.reference_files import spatialite_database, demo_init_py
from aequilibrae.transit.transit import Transit
from aequilibrae.utils.db_utils import commit_and_close
from aequilibrae.utils.model_run_utils import import_file_as_module


class Project:
    """AequilibraE project class

    .. code-block:: python
        :caption: Create Project

        >>> new_project = Project()
        >>> new_project.new(project_path)

    .. code-block:: python
        :caption: Open Project

        >>> existing_project = Project()
        >>> existing_project.open(project_path)
    """

    def __init__(self):
        self.path_to_file: str = None
        self.project_base_path = Path()
        self.source: str = None
        self.network: Network = None
        self.about: About = None
        self.logger: logging.Logger = None
        self.transit: Transit = None

    @classmethod
    def from_path(cls, project_folder):
        project = Project()
        project.open(project_folder)
        return project

    def open(self, project_path: str) -> None:
        """
        Loads project from disk

        :Arguments:
            **project_path** (:obj:`str`): Full path to the project data folder. If the project inside does
            not exist, it will fail.
        """

        self.project_base_path = Path(project_path)
        file_name = self.project_base_path / "project_database.sqlite"
        if not file_name.is_file() or not file_name.exists():
            raise FileNotFoundError("Model does not exist. Check your path and try again")
        self.path_to_file = file_name
        self.source = self.path_to_file
        self.__setup_logger()
        self.activate()

        self.__load_objects()
        global_logger.info(f"Opened project on {self.project_base_path}")
        clean(self)

    @property
    @contextmanager
    def db_connection(self):
        with commit_and_close(self.path_to_file, spatial=True) as conn:
            yield conn

    def new(self, project_path: str) -> None:
        """Creates a new project

        :Arguments:
            **project_path** (:obj:`str`): Full path to the project data folder. If folder exists, it will fail
        """

        self.project_base_path = Path(project_path)
        self.path_to_file = self.project_base_path / "project_database.sqlite"
        self.source = self.path_to_file

        if os.path.isdir(project_path):
            raise FileExistsError("Location already exists. Choose a different name or remove the existing directory")

        # We create the project folder and create the base file
        self.project_base_path.mkdir(parents=True, exist_ok=True)

        self.__setup_logger()
        self.activate()

        self.__create_empty_network()
        self.__load_objects()
        self.about.create()
        global_logger.info(f"Created project on {self.project_base_path}")

    def close(self) -> None:
        """Safely closes the project"""
        if not self.project_base_path:
            global_logger.warning("This Aequilibrae project is not opened")
            return

        try:
            with self.project.db_connection as conn:
                conn.commit()
            clean(self)
            for obj in [self.parameters, self.network]:
                del obj

            del self.network.link_types
            del self.network.modes

            global_logger.info(f"Closed project on {self.project_base_path}")

        except (sqlite3.ProgrammingError, AttributeError):
            global_logger.warning(f"This project at {self.project_base_path} is already closed")

        finally:
            self.deactivate()

    def activate(self):
        activate_project(self)

    def deactivate(self):
        if get_active_project(must_exist=False) is self:
            activate_project(None)

    def log(self) -> Log:
        """Returns a log object

        allows the user to read the log or clear it"""
        return Log(self.project_base_path)

    def upgrade(self):
        """
        Find and apply all applicable migrations.

        All database upgrades are applied within a single transaction.

        If skipping a specific migration is required, use the ``aequilbrae.project.tools.MigrationManager`` object
        directly. Consult it's documentation page for details. Take care when skipping migrations.
        """
        global_logger.info("Starting database upgrades")
        targets = [
            (MigrationManager(MigrationManager.network_migration_file), database_connection("project")),
        ]

        if (self.project_base_path / "public_transport.sqlite").exists():
            targets.append((MigrationManager(MigrationManager.transit_migration_file), database_connection("transit")))

        try:
            for mm, conn in targets:
                with conn:
                    mm.mark_all_as_seen(conn)

            for mm, conn in targets:
                with conn:
                    mm.upgrade(conn)
                    conn.execute("VACUUM")
            global_logger.info("Completed database upgrades")
        finally:
            for _, conn in targets:
                conn.close()

    def __load_objects(self):
        matrix_folder = self.project_base_path / "matrices"
        matrix_folder.mkdir(parents=True, exist_ok=True)

        self.network = Network(self)
        self.about = About(self)
        self.matrices = Matrices(self)

    @property
    def project_parameters(self) -> Parameters:
        return Parameters(self)

    @property
    def parameters(self) -> dict:
        return self.project_parameters.parameters

    @property
    def run(self):
        """
        Load and return the AequilibraE run module with the default arguments from
        ``parameters.yml`` partially applied.

        Refer to ``run/__init__.py`` file within the project folder for documentation.
        """
        entry_points = self.parameters["run"]
        module = import_file_as_module(self.project_base_path / "run" / "__init__.py", "aequilibrae.run")
        sentinal = object()
        for name, kwargs in entry_points.items():
            attr = getattr(module, name)
            if attr is sentinal:
                raise RuntimeError(f"expected to find callable '{name}' in the run module but didn't")
            elif not callable(attr):
                raise RuntimeError(f"found symbol '{name}' in the run module but it is not callable")

            func = functools.partial(attr, **kwargs)
            setattr(module, name, func)

        return module

    def check_file_indices(self) -> None:
        """Makes results_database.sqlite and the matrices folder compatible with project database"""
        raise NotImplementedError

    @property
    def zoning(self):
        return Zoning(self.network)

    def __create_empty_network(self):
        shutil.copyfile(spatialite_database, self.path_to_file)
        pth = self.project_base_path / "run"
        pth.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(demo_init_py, pth / "__init__.py")

        # Write parameters to the project folder
        p = self.project_parameters
        p.parameters["system"]["logging_directory"] = str(self.project_base_path)
        p.write_back()

        # Create actual tables
        with self.db_connection as conn:
            conn.execute("PRAGMA foreign_keys = ON;")
        initialize_tables(self, "network")

    def __setup_logger(self):
        self.logger = logging.getLogger(f"aequilibrae.{self.project_base_path}")
        self.logger.propagate = False
        self.logger.setLevel(logging.DEBUG)

        par = self.parameters or self.project_parameters._default
        do_log = par["system"]["logging"]

        if do_log:
            self.logger.addHandler(get_log_handler(self.project_base_path / "aequilibrae.log"))



import os
from typing import Dict, List
import numpy as np
import pandas as pd
from aequilibrae.context import get_active_project


# TODO: let's make it optional to keep path files in memory, although this can get out of control very quickly it should
# be much quicker when working with a single origin

# TODO: factor out AssignmentResultsTable into different file, also get rid of dirty eval bu changing creation

# FIXME: this is for zone_index and compressed link ids
# for link ids, look up what we are doing in graph - we might want to keep the order of links for a single compressed
# link
# for zones, we need to do the inverse of graph.compact_nodes_to_indices


class TrafficClassIdentifier(object):
    def __init__(self, name: str, id: str):
        self.name = name
        self.id = id


class AssignmentResultsTable(object):
    def __init__(self, table_name: str, project=None) -> None:
        self.project = project or get_active_project()
        self.table_name = table_name
        self.assignment_results = self._read_assignment_results()
        self.table_name = self.assignment_results["table_name"].values[0]
        self.procedure = self.assignment_results["procedure"].values[0]
        self.procedure_id = self.assignment_results["procedure_id"].values[0]
        self.timestamp = self.assignment_results["timestamp"].values[0]
        self.description = self.assignment_results["description"].values[0]
        self.procedure_report = self._parse_procedure_report()

    def _read_assignment_results(self) -> pd.DataFrame:
        with self.project.db_connection as conn:
            results_df = pd.read_sql("SELECT * FROM 'results'", conn)

        res = results_df.loc[results_df.table_name == self.table_name]
        assert len(res) == 1, f"Found {len(res)} assignment result with this table name, need exactly one"
        return res

    def _parse_procedure_report(self) -> Dict:
        rep_with_replacement = (
            self.assignment_results["procedure_report"].values[0].replace("inf", "np.inf").replace("nan", "np.nan")
        )
        report = eval(rep_with_replacement)
        report["convergence"] = eval(report["convergence"])
        report["setup"] = eval(report["setup"])
        return report

    def get_traffic_class_names_and_id(self) -> List[TrafficClassIdentifier]:
        all_classes = self.procedure_report["setup"]["Classes"]
        return [TrafficClassIdentifier(k, v["network mode"]) for k, v in all_classes.items()]


class AssignmentPaths(object):
    """Class for accessing path files optionally generated during assignment."""

    def __init__(self, table_name: str, project=None) -> None:
        """
        Instantiates the class

        :Arguments:
            **table_name** (str): Name of the traffic assignment result table used to generate the required path files

            **project** (:obj:`Project`, *Optional*): The Project to connect to.
            By default, uses the currently active project

        """
        project = project or get_active_project()
        self.proj_dir = project.project_base_path
        self.table_name = table_name
        self.assignment_results = AssignmentResultsTable(table_name, project)
        self.path_base_dir = os.path.join(self.proj_dir, "path_files", self.assignment_results.procedure_id)
        self.classes = self.assignment_results.get_traffic_class_names_and_id()
        self.compressed_graph_correspondences = self._read_compressed_graph_correspondence()

    def _read_compressed_graph_correspondence(self) -> Dict:
        compressed_graph_correspondences = {}
        for c in self.classes:
            compressed_graph_correspondences[c.id] = pd.read_feather(
                os.path.join(self.path_base_dir, f"correspondence_c{c.id}_{c.name}.feather")
            )
        return compressed_graph_correspondences

    def read_path_file(self, origin: int, iteration: int, traffic_class_id: str) -> (pd.DataFrame, pd.DataFrame):
        possible_traffic_classes = list(filter(lambda x: x.id == traffic_class_id, self.classes))
        class_ids = [x.id for x in self.classes]
        assert len(possible_traffic_classes) == 1, f"traffic class id not unique, please choose one of {class_ids}"
        traffic_class = possible_traffic_classes[0]
        b_dir = os.path.join(self.path_base_dir, f"iter{iteration}", f"path_c{traffic_class.id}_{traffic_class.name}")
        path_o_f = os.path.join(b_dir, f"o{origin}.feather")
        path_o_index_f = os.path.join(b_dir, f"o{origin}_indexdata.feather")
        path_o = pd.read_feather(path_o_f)
        path_o_index = pd.read_feather(path_o_index_f)
        return path_o, path_o_index

    def get_path_for_destination(self, origin: int, destination: int, iteration: int, traffic_class_id: str):
        """Return all link ids, i.e. the full path, for a given destination"""
        path_o, path_o_index = self.read_path_file(origin, iteration, traffic_class_id)
        return self.get_path_for_destination_from_files(path_o, path_o_index, destination)

    @staticmethod
    def get_path_for_destination_from_files(path_o: pd.DataFrame, path_o_index: pd.DataFrame, destination: int):
        """for a given path file and path index file, and a given destination, return the path links in o-d order"""
        if destination == 0:
            lower_incl = 0
        else:
            lower_incl = path_o_index.loc[path_o_index.index == destination - 1].values[0][0]
        upper_non_incl = path_o_index.loc[path_o_index.index == destination].values[0][0]
        links_on_path = path_o.loc[(path_o.index >= lower_incl) & (path_o.index < upper_non_incl)].values.flatten()
        links_on_path = np.flip(links_on_path)
        return links_on_path



import functools
import os
import tempfile
import uuid
import warnings
from copy import copy
from functools import reduce
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
import openmatrix as omx
from scipy.sparse import coo_matrix

# Checks if we can display OMX

# CONSTANTS
VERSION = 1  # VERSION OF THE MATRIX FORMAT
INT = 0
FLOAT = 1
COMPLEX = 2
CORE_NAME_MAX_LENGTH = 50
INDEX_NAME_MAX_LENGTH = 20
MATRIX_NAME_MAX_LENGTH = 20
MATRIX_DESCRIPTION_MAX_LENGTH = 144  # Why not a tweet?
NOT_COMPRESSED = 0
COMPRESSED = 1

# TODO:  Add an aggregate method
# TODO: Add check that the index 1 (or zero) has only unique values?


# Matrix structure
#
# What:  Version | Compress flag | # cells: compressed matrix | # of zones | # of cores | # of indices (Y) | Data type |
# Size:   uint8  |    uint8      |        uint64              |  uint32    |   uint8    |     uint8        |  uint8    |
# Shape:     1   |       1       |           1                |     1      |            |       1          |     1     |
# Offset:    0   |       1       |           2                |     10     |     14     |      15          |    16     |
#
#
# What:    Data size |  matrix name | matrix description | Core names |   index names    |
# Size:     uint8    |     S20      |          S144      |    S50     |      S20         |
# Shape:      1      |      1       |            1       |  [cores]   |    [indices]     |
# Offset:     17     |     18       |          38        |     182    |  182 + 50*cores  |
#
# What:         indices          |             Matrices                 |
# Size:         uint64           |      f(Data type, Data size)         |
# Shape:     [zones, indices]    |       [zones, zones, cores]          |
# Offset:  18 + 50*cores + Y*20  |   18 + 50*cores + Y*20 + Y*zones*8   |


matrix_export_types = ["Open matrix (*.omx), Aequilibrae matrix (*.aem)", "Comma-separated file (*.csv)"]


class AequilibraeMatrix(object):
    """
    Matrix class
    """

    def __init__(self):
        """
        Creates a memory instance for a matrix, that can be used to load an existing matrix or to create an empty one
        """
        self.file_path = None
        self.dtype = None
        self.num_indices = None
        self.index_names = None
        self.compressed = NOT_COMPRESSED
        self.matrix_view = None
        self.view_names = None
        self.matrix_hash = {}
        self.index = None
        self.indices = None
        self.matrix = None
        self.matrices = None
        self.cores = 0
        self.zones = None
        self.dtype = None
        self.names = []  # type: [str]
        self.name = ""
        self.description = ""
        self.current_index = None
        self.__omx = False
        self.__memory_only = True
        self.omx_file = None  # type: omx.File
        self.__version__ = VERSION  # Writes file version

    def save(self, names=(), file_name=None) -> None:
        """Saves matrix data back to file.

        If working with AEM file, it flushes data to disk. If working with OMX, requires new names.

        :Arguments:
            **names** (:obj:`tuple(str)`, *Optional*): New names for the matrices. Required if working with OMX files

            **file_name** (:obj:`str`, *Optional*): Local path to the matrix file
        """
        if file_name is not None:
            cores = names if len(names) else self.names
            self.__save_as(file_name, cores)
            return

        if not self.__omx:
            self.__flush(self.matrices)
            return

        if isinstance(names, str):
            names = [names]

        if len(names) != len(self.view_names):
            raise ValueError("Number of names needs to be equal to computational view")

        exists = [n for n in names if n in self.names]
        if exists:
            raise ValueError(f'Matrix(ces) "{".".join(exists)}" already exist. Choose (a) new name(s)')

        if len(self.view_names) == 1:
            self.omx_file[names[0]] = self.matrix_view[:, :]
        else:
            for i, name in enumerate(names):
                self.omx_file[name] = self.matrix_view[:, :, i]

        self.names = self.omx_file.list_matrices()
        self.computational_view(names)

    def __save_as(self, file_name: str, cores: List[str]):
        if Path(file_name).suffix.lower() == ".aem":
            mat = AequilibraeMatrix()
            args = {
                "zones": self.zones,
                "matrix_names": cores,
                "index_names": self.index_names,
                "memory_only": False,
                "file_name": file_name,
            }
            mat.create_empty(**args)
            mat.indices[:, :] = self.indices[:, :]
            for core in cores:
                mat.matrix[core][:, :] = self.matrix[core][:, :]
            mat.name = self.name
            mat.description = self.description
            mat.close()
            del mat

        elif Path(file_name).suffix.lower() == ".omx":
            omx_mat = omx.open_file(file_name, "w")
            for core in cores:
                omx_mat[core] = self.matrix[core]

            for index in self.index_names:
                omx_mat.create_mapping(index, self.indices[index])

            omx_mat.attrs.name = self.name
            omx_mat.attrs.description = self.description
            omx_mat.close()

    def create_empty(
        self,
        file_name: str = None,
        zones: int = None,
        matrix_names: List[str] = None,
        data_type: np.dtype = np.float64,
        index_names: List[str] = None,
        compressed: bool = False,
        memory_only: bool = True,
    ):
        """
        Creates an empty matrix in the AequilibraE format

        :Arguments:
            **file_name** (:obj:`str`): Local path to the matrix file

            **zones** (:obj:`int`): Number of zones in the model (Integer). Maximum number of zones in a matrix is
            4,294,967,296

            **matrix_names** (:obj:`list`): A regular Python list of names of the matrix. Limit is 50 characters each.
            Maximum number of cores per matrix is 256

            **data_type** (:obj:`np.dtype`, *Optional*): Data type of the matrix as NUMPY data types
            (``np.int32``, ``np.int64``, ``np.float32``, ``np.float64``). Defaults to ``np.float64``

            **index_names** (:obj:`list`, *Optional*):  A regular Python list of names for indices.
            Limit is 20 characters each. Maximum number of indices per matrix is 256

            **compressed** (:obj:`bool`, *Optional*): Whether it is a flat matrix or a compressed one
            (Boolean - Not yet implemented)

            **memory_only** (:obj:`bool`, *Optional*): Whether you want to keep the matrix copy in memory only.
            Defaults to ``True``

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeMatrix

            >>> zones_in_the_model = 3317
            >>> names_list = ['Car trips', 'pt trips', 'DRT trips', 'bike trips', 'walk trips']

            >>> mat = AequilibraeMatrix()
            >>> mat.create_empty(file_name=os.path.join(my_folder_path, 'my_matrix.aem'),
            ...                  zones=zones_in_the_model,
            ...                  matrix_names=names_list,
            ...                  memory_only=False)
            >>> mat.num_indices
            1
            >>> mat.zones
            3317
        """

        self.__memory_only = memory_only
        self.file_path = file_name
        self.zones = zones
        self.index_names = index_names
        self.dtype = data_type

        # Matrix compression still not supported
        if compressed:
            raise Warning("Matrix compression not yet supported")

        self.compressed = COMPRESSED if compressed else NOT_COMPRESSED

        if index_names is None:
            self.index_names = ["main_index"]
        else:
            if not isinstance(index_names, (list, tuple)):
                raise Exception("Index names need to be provided as a list")

            self.index_names = index_names
            for ind_name in index_names:
                if isinstance(ind_name, str):
                    if len(ind_name) > INDEX_NAME_MAX_LENGTH:
                        raise ValueError(
                            "Index names need to be be shorter than {}: {}".format(INDEX_NAME_MAX_LENGTH, ind_name)
                        )
                else:
                    raise ValueError("Index names need to be strings: " + str(ind_name))

        self.num_indices = len(self.index_names)

        if matrix_names is None:
            matrix_names = ["mat"]
        else:
            if isinstance(matrix_names, list) or isinstance(matrix_names, tuple):
                for mat_name in matrix_names:
                    if isinstance(mat_name, str):
                        if mat_name in object.__dict__:
                            raise ValueError(mat_name + " is a reserved name")
                        if len(mat_name) > CORE_NAME_MAX_LENGTH:
                            raise ValueError(f"Matrix names need to be shorter than {CORE_NAME_MAX_LENGTH}: {mat_name}")
                    else:
                        raise ValueError("Matrix core names need to be strings: " + str(mat_name))
            else:
                raise Exception("Matrix names need to be provided as a list")

        self.names = copy(matrix_names)
        self.cores = len(self.names)
        if self.zones is None:
            return
        if self.file_path is None and not memory_only:
            return
        self.__write__()

    def get_matrix(self, core: str, copy=False) -> np.ndarray:
        """Returns the data for a matrix core

        :Arguments:
            **core** (:obj:`str`): name of the matrix core to be returned

            **copy** (:obj:`bool`, *Optional*): return a copy of the data. Defaults to False

        :Returns:
            **object** (:obj:`np.ndarray`): NumPy array
        """
        if core not in self.names:
            raise AttributeError("Matrix core does not exist in this matrix")
        if self.__omx:
            return np.array(self.omx_file[core])
        else:
            mat = self.matrix[core]
            if copy:
                mat = np.copy(mat)
            return mat

    def create_from_omx(
        self,
        omx_path: str,
        file_path: str = None,
        cores: List[str] = None,
        mappings: List[str] = None,
        robust: bool = True,
        compressed: bool = False,
        memory_only: bool = True,
    ) -> None:
        """
        Creates an AequilibraeMatrix from an original OpenMatrix

        :Arguments:
            **omx_path** (:obj:`str`): Path to the OMX file one wants to import

            **file_path** (:obj:`str`, *Optional*): Path for the output AequilibraeMatrix

            **cores** (:obj:`list`, *Optional*): List of matrix cores to be imported

            **mappings** (:obj:`list`, *Optional*): List of the matrix mappings (i.e. indices, centroid numbers)
            to be imported

            **robust** (:obj:`bool`, *Optional*): Boolean for whether AequilibraE should try to adjust the names
            for cores and indices in case they are too long. Defaults to ``True``

            **compressed** (:obj:`bool`, *Optional*): Boolean for whether we should compress the output matrix.
            Not yet implemented

            **memory_only** (:obj:`bool`, *Optional*): Whether you want to keep the matrix copy in memory only. Defaults to ``True``

        """

        def robust_name(input_name: str, max_length: int, forbiden_names: List[str]) -> str:
            return_name = input_name
            if len(input_name) > max_length:
                return_name = input_name[:max_length]

            if return_name not in forbiden_names:
                return return_name

            for i in range(999999):
                x = len(str(i))

                if x + len(return_name) > max_length:
                    trial_name = return_name[: max_length - x] + str(i)
                else:
                    trial_name = return_name + str(i)

                if trial_name not in forbiden_names:
                    return trial_name

        if compressed:
            raise Warning("Matrix compression not yet supported")

        src = omx.open_file(omx_path, "a")
        try:
            avail_cores = src.list_matrices()

            if cores is None:
                do_cores = avail_cores
            else:
                do_cores = [x for x in cores if x in avail_cores]
                if cores != do_cores:
                    do_cores = [x for x in cores if x not in avail_cores]
                    raise ValueError("Cores listed not available in the OMX file: {}".format(do_cores))

            avail_idx = src.list_mappings()
            if mappings is None:
                do_idx = avail_idx
                if not avail_idx:
                    do_idx = ["zero_base_index"]
            else:
                do_idx = [x for x in mappings if x in avail_idx]
                if mappings != do_idx:
                    do_idx = [x for x in mappings if x not in avail_idx]
                    raise ValueError("Mappings/indices listed not available in the OMX file: {}".format(do_idx))

            shp = src.shape()
            if shp[0] != shp[1]:
                raise ValueError("AequilibraE only supports square matrices")
            zones = shp[0]

            if robust:
                # Use reduce as we have to keep track of which generated names have been used (to avoid collisions)
                core_names = functools.reduce(
                    lambda acc, n: acc + [robust_name(n, CORE_NAME_MAX_LENGTH, acc)], do_cores, []
                )
                idx_names = functools.reduce(
                    lambda acc, n: acc + [robust_name(n, INDEX_NAME_MAX_LENGTH, acc)], do_idx, []
                )
            else:
                core_names = list(do_cores)
                idx_names = list(do_idx)

            self.create_empty(
                file_name=file_path,
                zones=zones,
                matrix_names=core_names,
                index_names=idx_names,
                compressed=compressed,
                memory_only=memory_only,
            )

            # Copy all cores
            for ncore, core in zip(core_names, do_cores):
                self.matrix[ncore][:, :] = np.array(src[core])[:, :]
            self.__flush(self.matrices)

            # copy all indices
            if avail_idx:
                for nidx, idx in enumerate(do_idx):
                    ix = np.array(list(src.mapping(idx).keys()))
                    self.indices[:, nidx] = ix[:]
            else:
                self.index[:, 0] = np.arange(zones)

            self.__flush(self.indices)
        finally:
            src.close()

    def create_from_trip_list(self, path_to_file: str, from_column: str, to_column: str, list_cores: List[str]) -> str:
        """
        Creates an AequilibraeMatrix from a trip list csv file
        The output is saved in the same folder as the trip list file

        :Arguments:
            **path_to_file** (:obj:`str`): Path for the trip list csv file

            **from_column** (:obj:`str`): trip list file column containing the origin zones numbers

            **to_column** (:obj:`str`): trip list file column containing the destination zones numbers

            **list_cores** (:obj:`list`): list of core columns in the trip list file

        """

        # Loading file
        trip_df = pd.read_csv(path_to_file)

        # Creating zone indices
        zones_list = sorted(set(list(trip_df[from_column].unique()) + list(trip_df[to_column].unique())))
        zones_df = pd.DataFrame({"zone": zones_list, "idx": list(np.arange(len(zones_list)))})

        trip_df = trip_df.merge(
            zones_df.rename(columns={"zone": from_column, "idx": from_column + "_idx"}), on=from_column, how="left"
        ).merge(zones_df.rename(columns={"zone": to_column, "idx": to_column + "_idx"}), on=to_column, how="left")

        new_mat = AequilibraeMatrix()
        nb_of_zones = len(zones_list)
        new_mat.create_empty(
            file_name=path_to_file[:-4] + ".aem", zones=nb_of_zones, matrix_names=list_cores, memory_only=False
        )

        for idx, core in enumerate(list_cores):
            m = (
                coo_matrix(
                    (trip_df[core], (trip_df[from_column + "_idx"], trip_df[to_column + "_idx"])),
                    shape=(nb_of_zones, nb_of_zones),
                )
                .toarray()
                .astype(np.float64)
            )
            new_mat.matrix[new_mat.names[idx]][:, :] = m[:, :]

        new_mat.save()

        print(f"AequilibraE matrix saved at {path_to_file[:-4]}.aem")
        return

    def __load_aem__(self):
        # GET File version
        self.__version__ = np.memmap(self.file_path, dtype="uint8", offset=0, mode="r+", shape=1)[0]

        if self.__version__ != VERSION:
            raise ValueError("Matrix formats do not match")

        # If matrix is compressed or not
        self.compressed = np.memmap(self.file_path, dtype="uint8", offset=1, mode="r+", shape=1)[0]

        # number matrix cells if compressed
        _ = np.memmap(self.file_path, dtype="uint64", offset=2, mode="r+", shape=1)[0]

        # Zones
        self.zones = np.memmap(self.file_path, dtype="uint32", offset=10, mode="r+", shape=1)[0]

        # Matrix cores
        self.cores = np.memmap(self.file_path, dtype="uint8", offset=14, mode="r+", shape=1)[0]

        # Matrix indices
        self.num_indices = np.memmap(self.file_path, dtype="uint8", offset=15, mode="r+", shape=1)[0]

        # Data type
        data_class = np.memmap(self.file_path, dtype="uint8", offset=16, mode="r+", shape=1)[0]

        # Data size
        data_size = np.memmap(self.file_path, dtype="uint8", offset=17, mode="r+", shape=1)[0]

        if data_class == INT:
            if data_size == 1:
                self.dtype = np.int8
            elif data_size == 2:
                self.dtype = np.int16
            elif data_size == 4:
                self.dtype = np.int32
            elif data_size == 8:
                self.dtype = np.int64
            elif data_size == 16:
                self.dtype = np.int128

        if data_class == FLOAT:
            if data_size == 2:
                self.dtype = np.float16
            elif data_size == 4:
                self.dtype = np.float32
            elif data_size == 8:
                self.dtype = np.float64
            elif data_size == 16:
                self.dtype = np.float128

        # matrix name
        self.name = np.memmap(self.file_path, dtype="S" + str(MATRIX_NAME_MAX_LENGTH), offset=18, mode="r+", shape=1)[0]

        # matrix description
        offset = 18 + MATRIX_NAME_MAX_LENGTH
        self.description = np.memmap(
            self.file_path, dtype="S" + str(MATRIX_DESCRIPTION_MAX_LENGTH), offset=offset, mode="r+", shape=1
        )[0]

        # core names
        offset += MATRIX_DESCRIPTION_MAX_LENGTH
        self.names = list(
            np.memmap(self.file_path, dtype="S" + str(CORE_NAME_MAX_LENGTH), offset=offset, mode="r+", shape=self.cores)
        )

        self.names = [x.decode("utf-8") for x in self.names]

        # Index names
        offset += CORE_NAME_MAX_LENGTH * self.cores
        self.index_names = list(
            np.memmap(
                self.file_path, dtype="S" + str(INDEX_NAME_MAX_LENGTH), offset=offset, mode="r+", shape=self.num_indices
            )
        )
        self.index_names = [x.decode("utf-8") for x in self.index_names]

        # Index
        offset += self.num_indices * INDEX_NAME_MAX_LENGTH
        self.indices = np.memmap(
            self.file_path, dtype="uint64", offset=offset, mode="r+", shape=(self.zones, self.num_indices)
        )
        self.set_index(self.index_names[0])

        # DATA
        offset += self.zones * 8 * self.num_indices

        self.matrices = np.memmap(
            self.file_path, dtype=self.dtype, offset=offset, mode="r+", shape=(self.zones, self.zones, self.cores)
        )

        self.matrix = {}
        for i, v in enumerate(self.names):
            self.matrix[v] = self.matrices[:, :, i]
        self.matrix_hash = self.__builds_hash__()

    def __load_omx__(self):
        # GET File version
        self.__version__ = VERSION

        self.name = self.file_path
        self.description = "OMX MATRIX"
        self.names = self.omx_file.list_matrices()

        if len(self.omx_file) == 0:
            raise LookupError("Matrix file has no cores")

        self.index_names = self.omx_file.list_mappings()

        if len(self.index_names) == 0:
            raise LookupError("Matrix file has no indices (mappings). If you can't index it, you can't use it")

        self.num_indices = len(self.index_names)
        self.zones = len(list(self.omx_file.mapping(self.index_names[0]).keys()))

        self.cores = len(self.names)
        self.set_index(self.index_names[0])

        self.matrices = np.zeros(1)
        self.matrix = {}
        self.matrix_hash = self.__builds_hash__()

    def __write__(self):
        if not self.__memory_only:
            np.memmap(self.file_path, dtype="uint8", offset=0, mode="w+", shape=1)[0] = self.__version__

            # If matrix is compressed or not
            np.memmap(self.file_path, dtype="uint8", offset=1, mode="r+", shape=1)[0] = self.compressed

            # number matrix cells if compressed
            matrix_cells = self.zones * self.zones
            np.memmap(self.file_path, dtype="uint64", offset=2, mode="r+", shape=1)[0] = matrix_cells

            # Zones
            np.memmap(self.file_path, dtype="uint32", offset=10, mode="r+", shape=1)[0] = self.zones

            # Matrix cores
            np.memmap(self.file_path, dtype="uint8", offset=14, mode="r+", shape=1)[0] = self.cores

            # Matrix indices
            np.memmap(self.file_path, dtype="uint8", offset=15, mode="r+", shape=1)[0] = self.num_indices

            # Data type
            data_class = self.__define_data_class()
            np.memmap(self.file_path, dtype="uint8", offset=16, mode="r+", shape=1)[0] = data_class

            # Data size
            data_size = np.dtype(self.dtype).itemsize
            np.memmap(self.file_path, dtype="uint8", offset=17, mode="r+", shape=1)[0] = data_size

            # matrix name
            np.memmap(self.file_path, dtype=f"S{MATRIX_NAME_MAX_LENGTH}", offset=18, mode="r+", shape=1)[0] = self.name

            # matrix description
            offset = 18 + MATRIX_NAME_MAX_LENGTH
            np.memmap(
                self.file_path, dtype="S" + str(MATRIX_DESCRIPTION_MAX_LENGTH), offset=offset, mode="r+", shape=1
            )[0] = self.description

            # core names
            offset += MATRIX_DESCRIPTION_MAX_LENGTH
            fp = np.memmap(
                self.file_path, dtype="S" + str(CORE_NAME_MAX_LENGTH), offset=offset, mode="r+", shape=self.cores
            )

            for i, v in enumerate(self.names):
                fp[i] = v
            fp.flush()
            del fp

            # Index names
            offset += CORE_NAME_MAX_LENGTH * self.cores
            fp = np.memmap(
                self.file_path, dtype="S" + str(INDEX_NAME_MAX_LENGTH), offset=offset, mode="r+", shape=self.num_indices
            )

            for i, v in enumerate(self.index_names):
                fp[i] = v
            fp.flush()
            del fp

        # Index
        if self.__memory_only:
            self.indices = np.zeros((self.zones, self.num_indices), dtype=int)
        else:
            offset += self.num_indices * INDEX_NAME_MAX_LENGTH
            self.indices = np.memmap(
                self.file_path, dtype="uint64", offset=offset, mode="r+", shape=(self.zones, self.num_indices)
            )
            self.indices.fill(0)
            self.__flush(self.indices)
        self.set_index(self.index_names[0])

        if self.__memory_only:
            self.matrices = np.zeros((self.zones, self.zones, self.cores))
        else:
            offset += self.zones * 8 * self.num_indices
            self.matrices = np.memmap(
                self.file_path, dtype=self.dtype, offset=offset, mode="r+", shape=(self.zones, self.zones, self.cores)
            )

        if np.issubdtype(self.dtype, np.integer):
            self.matrices.fill(np.iinfo(self.dtype).min)
        else:
            self.matrices.fill(np.nan)

        self.__flush(self.matrices)

        self.matrix = {}
        for i, v in enumerate(self.names):
            self.matrix[v] = self.matrices[:, :, i]

    def set_index(self, index_to_set: str) -> None:
        """
        Sets the standard index to be the one the user wants to have be the one being used in all operations
        during run time. The first index is ALWAYS the default one every time the matrix is instantiated

        :Arguments:
            **index_to_set** (:obj:`str`): Name of the index to be used. The default index name is 'main_index'

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeMatrix

            >>> zones_in_the_model = 3317
            >>> names_list = ['Car trips', 'pt trips', 'DRT trips', 'bike trips', 'walk trips']
            >>> index_list = ['tazs', 'census']

            >>> mat = AequilibraeMatrix()
            >>> mat.create_empty(file_name=os.path.join(my_folder_path, 'my_matrix.aem'),
            ...                  zones=zones_in_the_model,
            ...                  matrix_names=names_list,
            ...                  index_names=index_list )
            >>> mat.num_indices
            2
            >>> mat.current_index
            'tazs'
            >>> mat.set_index('census')
            >>> mat.current_index
            'census'
        """
        if self.__omx:
            self.index = np.array(list(self.omx_file.mapping(index_to_set).keys()))
            self.current_index = index_to_set
        else:
            if index_to_set in self.index_names:
                ind_index = self.index_names.index(index_to_set)
                self.index = self.indices[:, ind_index]
                self.current_index = index_to_set
            else:
                raise ValueError("Index {} needs to be a string or its integer index.".format(str(index_to_set)))

    def __flush(self, obj_to_flush: np.memmap):
        if self.__memory_only:
            return
        obj_to_flush.flush()

    def __getattr__(self, mat_name: str):
        if mat_name in object.__dict__:
            if mat_name == "matrix" and self.__omx:
                warnings.warn("You can't access OMX matrix cores like that")
                return
            return self.__dict__[mat_name]

        if mat_name in self.names:
            return self.get_matrix(mat_name, copy=True)

        raise AttributeError("No such method or matrix core! --> " + str(mat_name))

    # Transforms matrix from dense to CSR
    # def compress(self):
    #     print(self.file_path + '. Method not implemented yet. All matrices are dense')

    # Transforms matrix from CSR to dense
    # def decompress(self):
    #     print(self.file_path + '. Method not implemented yet. All matrices are dense')

    # Adds index to matrix
    # def add_index(self):
    #     print(self.file_path + '. Method not implemented yet. All indices need to exist during the matrix creation')

    # Adds index to matrix
    # def remove_index(self, index_number):
    #     print(self.file_path + '. Method not implemented. Indices are fixed on matrix creation: ' + str(index_number))

    def close(self):
        """
        Removes matrix from memory and flushes all data to disk, or closes the OMX file if that is the case
        """

        if self.__omx:
            self.omx_file.close()
        else:
            self.__flush(self.matrices)
            self.__flush(self.index)

        for attr in ("index", "indices", "matrix", "matrices"):
            delattr(self, attr)

    def export(self, output_name: str, cores: List[str] = None):
        """
        Exports the matrix to other formats, rather than AEM. Formats currently supported: CSV, OMX

        When exporting to AEM or OMX, the user can chose to export only a set of cores, but all indices are exported

        When exporting to CSV, the active index will be used, and all cores will be exported as separate columns in
        the output file

        :Arguments:
            **output_name** (:obj:`str`): Path to the output file

            **cores** (:obj:`list`): Names of the cores to be exported.

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeMatrix

            >>> zones_in_the_model = 3317
            >>> names_list = ['Car trips', 'pt trips', 'DRT trips', 'bike trips', 'walk trips']

            >>> mat = AequilibraeMatrix()
            >>> mat.create_empty(file_name=os.path.join(my_folder_path, 'my_matrix.aem'),
            ...                  zones=zones_in_the_model,
            ...                  matrix_names=names_list)

            >>> mat.export(os.path.join(my_folder_path, 'my_new_path.aem'), ['Car trips', 'bike trips'])

            >>> mat2 = AequilibraeMatrix()
            >>> mat2.load(os.path.join(my_folder_path, 'my_new_path.aem'))
            >>> mat2.cores
            2
        """
        fname, file_extension = os.path.splitext(output_name.upper())

        if file_extension not in [".AEM", ".CSV", ".OMX"]:
            raise NotImplementedError(f"File extension {file_extension} not implemented yet")

        if cores is None:
            cores = self.names

        if file_extension == ".AEM":
            self.copy(output_name=output_name, cores=cores, memory_only=False)

        elif file_extension == ".OMX":
            omx_export = omx.open_file(output_name, "w")
            for c in cores:
                if self.__omx:
                    omx_export[c] = np.array(self.omx_file[c])
                else:
                    omx_export[c] = self.matrix[c]
            for i, idx in enumerate(self.index_names):
                omx_export.create_mapping(idx, self.indices[:, i])
            omx_export.close()

        elif file_extension == ".CSV":

            def f(name):
                if self.__omx:
                    coo = coo_matrix(np.array(self.omx_file[name]))
                else:
                    coo = coo_matrix(self.matrix[name])
                data = {"row": self.index[coo.row], "column": self.index[coo.col], name: coo.data}
                return pd.DataFrame(data).set_index(["row", "column"])

            dfs = [f(name) for name in self.names]
            df = reduce(lambda a, b: a.join(b, how="outer"), dfs)
            df.to_csv(output_name, index=True)

    def load(self, file_path: str):
        """
        Loads matrix from disk. All cores and indices are load. First index is default.

        :Arguments:
            **file_path** (:obj:`str`): Path to AEM or OMX file on disk

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeMatrix

            >>> project = create_example(project_path)

            >>> mat = AequilibraeMatrix()
            >>> mat.load(os.path.join(project_path, 'matrices/skims.omx'))
            >>> mat.computational_view()
            >>> mat.names
            ['distance_blended', 'time_final']
        """

        self.file_path = file_path

        if os.path.splitext(file_path)[-1].upper() == ".OMX":
            self.__omx = True
            self.omx_file = omx.open_file(file_path, "a")
            self.__load_omx__()
        else:
            self.__load_aem__()

    def is_omx(self):
        """Returns ``True`` if matrix data source is OMX, ``False`` otherwise"""
        return self.__omx

    def computational_view(self, core_list: List[str] = None):
        """
        Creates a memory view for a list of matrices that is compatible with Cython memory buffers

        It allows for AequilibraE matrices to be used in all parallelized algorithms within AequilibraE

        In case of OMX matrices, the computational view is held only in memory

        :Arguments:
            **core_list** (:obj:`list`): List with the names of all matrices that need to be in the buffer

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeMatrix

            >>> zones_in_the_model = 3317
            >>> names_list = ['Car trips', 'pt trips', 'DRT trips', 'bike trips', 'walk trips']

            >>> mat = AequilibraeMatrix()
            >>> mat.create_empty(file_name=os.path.join(my_folder_path, 'my_matrix.aem'),
            ...                  zones=zones_in_the_model,
            ...                  matrix_names=names_list)
            >>> mat.computational_view(['bike trips', 'walk trips'])
            >>> mat.view_names
            ['bike trips', 'walk trips']
        """

        self.matrix_view = None
        self.view_names = None
        core_list = core_list or self.names
        if isinstance(core_list, str):
            core_list = [core_list]

        missing = [i for i in core_list if i not in self.names]
        if len(missing) > 0:
            raise ValueError(f"Matrix core(s) {','.join(missing)} not available on this matrix")

        if self.__omx:
            self.view_names = core_list
            if len(core_list) == 1:
                self.matrix_view = np.array(self.omx_file[core_list[0]])
            else:
                self.matrix_view = np.empty((self.zones, self.zones, len(core_list)))
                for i, core in enumerate(core_list):
                    self.matrix_view[:, :, i] = np.array(self.omx_file[core])
        else:
            # Check if matrices are adjacent
            if len(core_list) > 1:
                for i, x in enumerate(core_list[1:]):
                    k = self.names.index(x)  # index of the first element
                    k0 = self.names.index(core_list[i])  # index of the first element
                    if k - k0 != 1:
                        raise ValueError("Matrix cores {} and {} are not adjacent".format(core_list[i - 1], x))

            self.view_names = core_list
            idx1 = self.names.index(core_list[0])
            if len(core_list) == 1:
                self.matrix_view = self.matrices[:, :, idx1]
            elif len(core_list) > 1:
                idx2 = self.names.index(core_list[-1])
                self.matrix_view = self.matrices[:, :, idx1 : idx2 + 1]

    def copy(
        self,
        output_name: str = None,
        cores: List[str] = None,
        names: List[str] = None,
        compress: bool = None,
        memory_only: bool = True,
    ):
        """
        Copies a list of cores (or all cores) from one matrix file to another one

        :Arguments:
            **output_name** (:obj:`str`): Name of the new matrix file.
            If none is provided, returns a copy in memory only

            **cores** (:obj:`list`): List of the matrix cores to be copied

            **names** (:obj:`list`, *Optional*): List with the new names for the cores. Defaults to current names

            **compress** (:obj:`bool`, *Optional*): Whether you want to compress the matrix or not.
            Defaults to ``False``. Not yet implemented

            **memory_only** (:obj:`bool`, *Optional*): Whether you want to keep the matrix copy in memory only.
            Defaults to ``True``

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeMatrix

            >>> zones_in_the_model = 3317
            >>> names_list = ['Car trips', 'pt trips', 'DRT trips', 'bike trips', 'walk trips']

            >>> mat = AequilibraeMatrix()
            >>> mat.create_empty(file_name=os.path.join(my_folder_path, 'my_matrix.aem'),
            ...                  zones=zones_in_the_model,
            ...                  matrix_names=names_list)

            >>> mat.copy(os.path.join(my_folder_path, 'copy_of_my_matrix.aem'),
            ...          cores=['bike trips', 'walk trips'],
            ...          names=['bicycle', 'walking'],
            ...          memory_only=False)  # doctest: +ELLIPSIS
            <aequilibrae.matrix.aequilibrae_matrix.AequilibraeMatrix object at 0x...>

            >>> mat2 = AequilibraeMatrix()
            >>> mat2.load(os.path.join(my_folder_path, 'copy_of_my_matrix.aem'))
            >>> mat2.cores
            2
        """
        if compress:
            raise Warning("Matrix compression not yet supported")

        if output_name is None:
            memory_only = True

        mcores = cores or self.names
        mnames = names or mcores

        if len(mcores) != len(mnames):
            raise ValueError("Must have same number of cores to copy and output names")

        orig_mat_view = self.view_names or self.names

        output = AequilibraeMatrix()
        output.create_empty(
            file_name=output_name,
            zones=self.zones,
            matrix_names=mnames,
            memory_only=memory_only,
            index_names=self.index_names,
        )

        for i, core in enumerate(mcores):
            self.computational_view([core])
            output.matrices[:, :, i] = self.matrix_view[:, :]

        for i, inm in enumerate(self.index_names):
            self.set_index(inm)
            output.indices[:, i] = self.index[:]
        output.matrix_hash = output.__builds_hash__()
        output.computational_view(mnames)
        if not memory_only:
            output.matrices.flush()
        self.computational_view(orig_mat_view)
        output.computational_view([core for core, old_name in zip(mnames, mcores) if old_name in orig_mat_view])
        return output

    def rows(self) -> np.ndarray:
        """
        Returns row vector for the matrix in the computational view

        Computational view needs to be set to a single matrix core

        :Returns:

            **object** (:obj:`np.ndarray`): the row totals for the matrix currently on the computational view

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeMatrix

            >>> project = create_example(project_path)

            >>> mat = AequilibraeMatrix()
            >>> mat.load(os.path.join(project_path, 'matrices/skims.omx'))
            >>> mat.computational_view(["distance_blended"])
            >>> mat.rows()
            array([357.68202084, 358.68778868, 310.68285491, 275.87964738,
                   265.91709918, 268.60184371, 267.32264726, 281.3793747 ,
                   286.15085073, 242.60308705, 252.1776242 , 305.56774194,
                   303.58100777, 270.48841269, 263.20417379, 253.92665702,
                   277.1655432 , 258.84368258, 280.65697316, 272.7651157 ,
                   264.06806038, 252.87533845, 273.45639965, 281.61102767])
        """
        return self.__vector(axis=0)

    def columns(self) -> np.ndarray:
        """
        Returns column vector for the matrix in the computational view

        Computational view needs to be set to a single matrix core

        :Returns:

            **object** (:obj:`np.ndarray`): the column totals for the matrix currently on the computational view

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeMatrix

            >>> project = create_example(project_path)

            >>> mat = AequilibraeMatrix()
            >>> mat.load(os.path.join(project_path, 'matrices/skims.omx'))
            >>> mat.computational_view(["distance_blended"])
            >>> mat.columns()
            array([357.54256811, 357.45109051, 310.88655449, 276.6783439 ,
                   266.70388637, 270.62976319, 266.32888632, 279.6897402 ,
                   285.89821842, 242.79743295, 252.34085912, 301.78116548,
                   302.97058146, 270.61855294, 264.59944248, 257.83842251,
                   276.63310578, 257.74513863, 281.15724257, 271.63886077,
                   264.62215032, 252.79791125, 273.18139747, 282.7636574 ])
        """
        return self.__vector(axis=1)

    def nan_to_num(self):
        """
        Converts all ``NaN`` values in all cores in the computational view to zeros

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeMatrix

            >>> nan_matrix = np.empty((3,3))
            >>> nan_matrix[:] = np.nan

            >>> index = np.arange(1, 4, dtype=np.int32)

            >>> mat = AequilibraeMatrix()
            >>> mat.create_empty(file_name=os.path.join(my_folder_path, "matrices/nan_matrix.aem"),
            ...                  zones=3,
            ...                  matrix_names=["only_nan"])
            >>> mat.index[:] = index[:]
            >>> mat.matrix["only_nan"][:, :] = nan_matrix[:, :]
            >>> mat.computational_view()
            >>> mat.nan_to_num()
            >>> mat.get_matrix("only_nan")
            array([[0., 0., 0.],
                   [0., 0., 0.],
                   [0., 0., 0.]])
        """

        if self.__omx:
            raise NotImplementedError("This operation does not make sense for OMX matrices")

        if np.issubdtype(self.dtype, np.floating) and self.matrix_view is not None:
            self.matrices[:, :, :] = np.nan_to_num(self.matrices[:, :, :])

    def __vector(self, axis: int):
        if self.view_names is None:
            raise ReferenceError("Matrix is not set for computation")
        if len(self.view_names) > 1:
            raise ValueError("Vector for a multi-core matrix is ambiguous")
        return self.matrix_view.astype(float).sum(axis=axis)[:]

    def __builds_hash__(self):
        return {self.index[i]: i for i in range(self.zones)}

    def __define_data_class(self):
        if self.__omx:
            raise NotImplementedError("This operation does not make sense for OMX matrices")

        if np.issubdtype(self.dtype, np.floating):
            data_class = FLOAT
        elif np.issubdtype(self.dtype, np.integer):
            data_class = INT
        else:
            raise ValueError(
                "Data type not supported. You can choose Integers of 8, 16, 32 and 64 bits, "
                "or floats with 16, 32 or 64 bits"
            )
        return data_class

    def setName(self, matrix_name: str):
        """
        Sets the name for the matrix itself. Only works for matrices in disk.

        :Arguments:
            **matrix_name** (:obj:`str`): matrix name. Maximum length is 50 characters

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeMatrix

            >>> zones_in_the_model = 3317

            >>> mat = AequilibraeMatrix()
            >>> mat.create_empty(file_name=os.path.join(my_folder_path, 'my_matrix.aem'),
            ...                  zones=zones_in_the_model,
            ...                  memory_only=False)
            >>> mat.setName('This is my example')
            >>> mat.save()
            >>> mat.close()

            >>> mat = AequilibraeMatrix()
            >>> mat.load(os.path.join(my_folder_path, 'my_matrix.aem'))
            >>> mat.name.decode('utf-8')
            'This is my example'
        """
        if self.__omx:
            raise NotImplementedError("This operation does not make sense for OMX matrices")

        if self.__memory_only:
            raise NotImplementedError("This operation does not make sense for memory only matrices")

        if matrix_name is not None:
            if len(str(matrix_name)) > MATRIX_NAME_MAX_LENGTH:
                matrix_name = str(matrix_name)[0:MATRIX_NAME_MAX_LENGTH]

            np.memmap(self.file_path, dtype=f"S{MATRIX_NAME_MAX_LENGTH}", offset=18, shape=1)[0] = matrix_name

    def setDescription(self, matrix_description: str):
        """
        Sets description for the matrix

        :Arguments:
            **matrix_description** (:obj:`str`): Text with matrix description. Maximum length is 144 characters

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeMatrix

            >>> zones_in_the_model = 3317

            >>> mat = AequilibraeMatrix()
            >>> mat.create_empty(file_name=os.path.join(my_folder_path, 'my_matrix.aem'),
            ...                  zones=zones_in_the_model,
            ...                  memory_only=False)
            >>> mat.setDescription('This is a text')
            >>> mat.save()
            >>> mat.close()

            >>> mat = AequilibraeMatrix()
            >>> mat.load(os.path.join(my_folder_path, 'my_matrix.aem'))
            >>> mat.description.decode('utf-8')
            'This is a text'
        """
        if self.__omx:
            raise NotImplementedError("This operation does not make sense for OMX matrices")

        if self.__memory_only:
            raise NotImplementedError("This operation does not make sense for memory only matrices")

        if matrix_description is not None:
            if len(str(matrix_description)) > MATRIX_DESCRIPTION_MAX_LENGTH:
                matrix_description = str(matrix_description)[0:MATRIX_DESCRIPTION_MAX_LENGTH]

            np.memmap(
                self.file_path,
                dtype="S" + str(MATRIX_NAME_MAX_LENGTH),
                offset=18 + MATRIX_NAME_MAX_LENGTH,
                mode="r+",
                shape=1,
            )[0] = matrix_description

    @staticmethod
    def random_name() -> str:
        """
        Returns a random name for a matrix with root in the temp directory of the user

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeMatrix

            >>> mat = AequilibraeMatrix()
            >>> mat.random_name() # doctest: +ELLIPSIS
            '/tmp/Aequilibrae_matrix_...'
        """
        return os.path.join(tempfile.gettempdir(), f"Aequilibrae_matrix_{uuid.uuid4()}.aem")



import sys
import pathlib
import importlib.util


def import_file_as_module(file: pathlib.Path, module_name):
    """
    Import a file as a Python module.

    :Arguments:
        **file** (:obj:`pathlib.Path`): Path object pointing to the file to import

        **module_name**: Name to give the imported module

    :Returns:
        The imported module
    """
    spec = importlib.util.spec_from_file_location(module_name, file)
    if spec is None:
        raise ImportError(f"Could not find module spec for {file}")

    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module



from sqlite3 import Connection
from aequilibrae.transit.constants import Constants, AGENCY_MULTIPLIER


class Fare:
    """Transit Fare

    * fare_id (:obj:`int`): ID of the fare as in the network model
    * fare (:obj:`str`): ID of the fare as in GTFS
    * agency (:obj:`str`): Corresponding agency as inputed during import
    * agency_id (:obj:`int`): ID of the corresponding agency as in the network model
    * price (:obj:`int`): As in GTFS
    * currency (:obj:`str`): As in GTFS
    * payment_method (:obj:`int`): As in GTFS
    * transfer (:obj:`int`): As in GTFS
    * transfer_duration (:obj:`int`): As in GTFS"""

    def __init__(self, agency_id: int):
        self.fare = ""
        self.fare_id = -1
        self.agency = ""
        self.agency_id = agency_id
        self.price = -1
        self.currency = ""
        self.payment_method = 0
        self.transfer = 0
        self.transfer_duration = 0
        self.__get_fare_id()

    def populate(self, record: tuple, headers: list) -> None:
        """Adds fare information."""
        for key, value in zip(headers, record):
            if key not in self.__dict__.keys():
                raise KeyError(f"{key} field in Routes.txt is unknown field for that file on GTFS")

            # We convert route_id into route, as the the GTFS header for it is not maintained in our database
            v = None if value in [0, ""] else value
            key = "fare" if key == "fare_id" else key
            key = "agency" if key == "agency_id" else key
            if key in ["payment_method", "transfer", "transfer_duration"]:
                v = v or self.__dict__[key]
            self.__dict__[key] = v

    def save_to_database(self, conn: Connection) -> None:
        """Saves Fare attributes to the database"""

        data = [
            self.fare_id,
            self.fare,
            self.agency_id,
            self.price,
            self.currency,
            self.payment_method,
            self.transfer,
            self.transfer_duration,
        ]
        sql = """insert into fare_attributes (fare_id, fare, agency_id, price, currency, payment_method,
                                                     transfer, transfer_duration) VALUES (?, ?, ?, ?, ?, ?, ?, ?);"""
        conn.execute(sql, data)
        conn.commit()

    def __get_fare_id(self):
        c = Constants()
        self.fare_id = 1 + c.fares.get(self.agency_id, AGENCY_MULTIPLIER * self.agency_id)
        c.fares[self.agency_id] = self.fare_id



from aequilibrae.utils.qgis_utils import inside_qgis


def noop(_):
    pass


if inside_qgis:
    from qgis.PyQt.QtCore import pyqtSignal as SIGNAL  # type: ignore

    noop(SIGNAL.__class__)  # This should be no-op but it stops PyCharm from "optimising" the above import
else:
    from aequilibrae.utils.python_signal import PythonSignal as SIGNAL  # type: ignore

    noop(SIGNAL.__class__)  # This should be no-op but it stops PyCharm from "optimising" the above import


class simple_progress(object):
    """A `tqdm` style iterable wrapper using aequilibrae signals"""

    def __init__(self, thing, signal, msg=None):
        self.thing = thing
        self.iterator = None

        try:
            num_elements = len(self.thing)
        except (TypeError, AttributeError):
            num_elements = 0

        self.msg = msg or f"{{}}/{num_elements}"
        self.signal = signal
        self.signal.emit(["start", num_elements, self.msg.format(0)])
        self.counter = 1

    def __iter__(self):
        self.iterator = iter(self.thing)
        return self

    def __next__(self):
        current = next(self.iterator)
        self.signal.emit(["update", self.counter, self.msg.format(self.counter)])
        self.counter = self.counter + 1
        return current
